import logging
import os
import tempfile


from datetime import datetime
from pathlib import Path
from snakemake.cli import args_to_api as snakemake_cli, parse_args
from typing import TextIO, List, Optional

from omni_schema.datamodel.omni_schema import SoftwareBackendEnum

from omnibenchmark.benchmark import Benchmark, BenchmarkNode
from omnibenchmark.workflow.workflow import WorkflowEngine
from omnibenchmark.workflow.snakemake import rules


# Module includes for each Snakefile
INCLUDES = [
    "utils.smk",
    "rule_start_benchmark.smk",
    "rule_node.smk",
    "rule_all.smk",
]


class SnakemakeEngine(WorkflowEngine):
    """Snakemake implementation of the WorkflowEngine interface."""

    def run_workflow(
        self,
        benchmark: Benchmark,
        cores: int = 1,
        update: bool = True,
        dryrun: bool = False,
        continue_on_error: bool = False,
        keep_module_logs: bool = False,
        backend: SoftwareBackendEnum = SoftwareBackendEnum.host,
        module_path: str = os.environ.get("MODULEPATH", ""),
        debug: bool = False,
        work_dir: Path = Path(os.getcwd()),
        out_dir: str = "out",
        local_timeout: Optional[int] = None,
        **snakemake_kwargs,
    ) -> bool:
        """
        Serializes & runs benchmark workflow using snakemake.

        Args:
            benchmark (Benchmark): benchmark to run
            cores (int): number of cores to run. Defaults to 1 core.
            update (bool): run workflow for non-existing outputs / changed nodes only. False means force running workflow from scratch. Default: True
            dryrun (bool): validate the workflow with the benchmark without actual execution. Default: False
            continue_on_error (bool): continue with independent jobs if a job fails. Default: False
            keep_module_logs (bool): keep module-specific log files after execution. Default: False
            backend (SoftwareBackendEnum): which software backend to use when running the workflow. Available: `host`, `docker`, `apptainer`, `conda`, `envmodules`. Default: `host`
            module_path (str): The path where the `envmodules` are located. This path will be searched during the workflow run using `envmodules` backend.
            work_dir (str): working directory. Default: current work directory
            out_dir (str): output directory. Default: `out`

            **snakemake_kwargs: keyword arguments to pass to the snakemake engine

        Returns:
        - Status code (bool) of the workflow run.
        """

        # Serialize Snakefile for workflow
        snakefile = self.serialize_workflow(
            benchmark, work_dir, write_to_disk=False, local_timeout=local_timeout
        )

        # Prepare the argv list
        argv = self._prepare_argv(
            snakefile=snakefile,
            cores=cores,
            update=update,
            dryrun=dryrun,
            continue_on_error=continue_on_error,
            keep_module_logs=keep_module_logs,
            backend=backend,
            work_dir=work_dir,
            debug=debug,
            out_dir=out_dir,
            **snakemake_kwargs,
        )

        if module_path:
            os.environ["MODULEPATH"] = module_path

        logging.getLogger("snakemake").setLevel(logging.DEBUG)

        # Execute snakemake script
        parser, args = parse_args(argv)
        return snakemake_cli(args, parser)

    def serialize_workflow(
        self,
        benchmark: Benchmark,
        output_dir: Path = Path(os.getcwd()),
        write_to_disk=True,
        local_timeout=1000,
    ) -> Path:
        """
        Serializes a Snakefile for the benchmark.

        Args:
            benchmark (Benchmark): benchmark to serialize
            output_dir (str): output directory for the Snakefile
            write_to_disk (bool): if write_to_disk is True, create Snakefile on disk, else create an in-memory temp file

        Returns:
        - Snakefile path.
        """
        os.makedirs(output_dir, exist_ok=True)

        benchmark_file = benchmark.get_definition_file()
        name = benchmark.get_benchmark_name()
        version = benchmark.get_benchmark_version()
        author = benchmark.get_benchmark_author()

        if write_to_disk:
            snakefile_path = Path(os.path.join(output_dir, "Snakefile"))
        else:
            temp_file = tempfile.NamedTemporaryFile(
                delete=False, mode="w+", suffix=".smk", encoding="utf-8"
            )
            snakefile_path = Path(temp_file.name)

        # Serialize Snakemake file
        with open(snakefile_path, "w", encoding="utf-8") as f:
            self._write_snakefile_header(f, name, version, author)
            self._write_includes(f, INCLUDES)

            # Load benchmark from yaml file
            f.write(f'benchmark = load("{benchmark_file.as_posix()}", config)\n\n')

            # Create capture all rule
            f.write("node_paths = sorted(benchmark.get_output_paths())\n")
            f.write(
                "mc_paths = sorted(benchmark.get_metric_collector_output_paths())\n"
            )
            f.write("all_paths = node_paths + mc_paths\n")
            f.write("create_all_rule(all_paths, True)\n\n")

            # Create node rules
            f.write("nodes = benchmark.get_nodes()\n")
            f.write("for node in nodes:\n")
            f.write(
                f"    create_node_rule(node, benchmark, config, {local_timeout})\n\n"
            )

            # Create metric collector rules
            f.write("collectors = benchmark.get_metric_collectors()\n")
            f.write("for collector in collectors:\n")
            f.write(
                "    create_metric_collector_rule(benchmark, collector, node_paths)\n\n"
            )

        return snakefile_path

    def run_node_workflow(
        self,
        node: BenchmarkNode,
        input_dir: Path,
        dataset: str,
        cores: int = 1,
        update: bool = True,
        dryrun: bool = False,
        continue_on_error: bool = False,
        keep_module_logs: bool = False,
        backend: SoftwareBackendEnum = SoftwareBackendEnum.host,
        module_path: str = os.environ.get("MODULEPATH", ""),
        work_dir: Path = Path(os.getcwd()),
        **snakemake_kwargs,
    ) -> bool:
        """
        Serializes & runs benchmark node workflow using snakemake.

        Args:
            node (Benchmark): benchmark node to run
            input_dir (str): directory containing the inputs for the benchmark node
            dataset (str): file names corresponding to the dataset
            cores (int): number of cores to run. Defaults to 1 core.
            update (bool): run workflow for non-existing outputs / changed nodes only. False means force running workflow from scratch. Default: True
            dryrun (bool): validate the workflow with the benchmark without actual execution. Default: False
            continue_on_error (bool): continue with independent jobs if a job fails. Default: False
            keep_module_logs (bool): keep module-specific log files after execution. Default: False
            backend (SoftwareBackendEnum): which software backend to use when running the workflow. Available: `host`, `docker`, `apptainer`, `conda`, `envmodules`. Default: `host`
            module_path (str): The path where the `envmodules` are located. This path will be searched during the workflow run using `envmodules` backend.
            work_dir (str): working directory. Default: current work directory
            **snakemake_kwargs: keyword arguments to pass to the snakemake engine

        Returns:
        - Status code of the workflow run.
        """

        os.makedirs(work_dir, exist_ok=True)

        # Serialize Snakefile for node workflow
        snakefile = self.serialize_node_workflow(node, work_dir, write_to_disk=True)

        # Prepare the argv list
        argv = self._prepare_argv(
            snakefile=snakefile,
            cores=cores,
            update=update,
            dryrun=dryrun,
            continue_on_error=continue_on_error,
            keep_module_logs=keep_module_logs,
            backend=backend,
            work_dir=work_dir,
            input_dir=input_dir,
            dataset=dataset,
            **snakemake_kwargs,
        )

        if module_path:
            os.environ["MODULEPATH"] = module_path

        # Execute snakemake script
        parser, args = parse_args(argv)

        success = snakemake_cli(args, parser)

        return success

    def serialize_node_workflow(
        self,
        node: BenchmarkNode,
        output_dir: Path = Path(os.getcwd()),
        write_to_disk=True,
    ) -> Path:
        """
        Serializes a Snakefile for a benchmark node.

        Args:
            node (BenchmarkNode): benchmark node to serialize
            output_dir (str): output directory for the Snakefile
            write_to_disk (bool): if write_to_disk is True, create Snakefile on disk, else create an in-memory temp file

        Returns:
        - Snakefile path.
        """
        os.makedirs(output_dir, exist_ok=True)

        benchmark_file = node.get_definition_file()
        name = node.get_benchmark_name()
        version = node.get_benchmark_version()
        author = node.get_benchmark_author()

        if write_to_disk:
            snakefile_path = Path(os.path.join(output_dir, "Snakefile"))
        else:
            temp_file = tempfile.NamedTemporaryFile(
                delete=False, mode="w+", suffix=".smk", encoding="utf-8"
            )
            snakefile_path = Path(temp_file.name)

        # Serialize Snakemake file
        with open(snakefile_path, "w") as f:
            self._write_snakefile_header(f, name, version, author)
            self._write_includes(f, INCLUDES)

            # Load benchmark from yaml file
            f.write(
                f'node = load_node("{benchmark_file.as_posix()}", "{node.get_id()}")\n\n'
            )

            # Create capture all rule
            f.write("input_paths = node.get_input_paths(config)\n")
            f.write("output_paths = node.get_output_paths(config)\n")
            f.write("all_paths = input_paths + output_paths\n\n")
            f.write("create_all_rule(all_paths)\n\n")

            # Create node rules
            f.write("create_standalone_node_rule(node, config)\n\n")

        return snakefile_path

    @staticmethod
    def _write_snakefile_header(
        f: TextIO, benchmark_name: str, benchmark_version: str, benchmark_author: str
    ):
        """Write header for the generated Snakefile"""

        f.write("#!/usr/bin/env snakemake -s\n")
        f.write("#############################################\n")
        f.write("# Snakefile for Orchestrating YAML-defined OmniBenchmarks\n")
        f.write("#############################################\n")
        f.write("# \n")
        f.write(
            f"# This Snakefile has been automatically generated on {datetime.now()}\n"
        )
        f.write("# \n")
        f.write("# Benchmark Details:\n")
        f.write(f"# - Name: {benchmark_name}\n")
        f.write(f"# - Version: {benchmark_version}\n")
        f.write("# \n")
        f.write("# Author Information:\n")
        f.write(f"# - Contact: {benchmark_author}\n")
        f.write("#\n")
        f.write("#############################################\n")
        f.write("\n")

    @staticmethod
    def _write_includes(f: TextIO, includes: List[str]):
        """Write includes directive for the generated Snakefile"""

        includes_path = Path(rules.__file__).resolve().parent
        for include in includes:
            include_path = includes_path / include
            f.write(f'include: "{include_path.as_posix()}"\n')

        f.write("\n")

    @staticmethod
    def _prepare_argv(
        snakefile: Path,
        cores: int,
        update: bool,
        dryrun: bool,
        keep_module_logs: bool,
        continue_on_error: bool,
        backend: SoftwareBackendEnum,
        work_dir: Path,
        input_dir: Optional[Path] = None,
        dataset: Optional[str] = None,
        debug: bool = False,
        out_dir: Optional[str] = None,
        **snakemake_kwargs,
    ):
        """Prepare arguments to input to the snakemake cli"""
        argv = [
            "--snakefile",
            str(snakefile),
            "--cores",
            str(cores),
            "--directory",
            work_dir.as_posix(),
            "--config",
            f"dataset={dataset}",
            f"keep_module_logs={keep_module_logs}",
            f"keep_going={continue_on_error}",
        ]

        if input_dir:
            argv.extend([f"input={input_dir.as_posix()}"])

        if out_dir:
            argv.extend([f"out_dir={out_dir}"])

        if debug:
            argv.extend(["--verbose", "--debug"])

        if update:
            argv.append("-F")

        if dryrun:
            argv.append("--dryrun")

        if continue_on_error:
            argv.append("--keep-going")

        if (
            backend == SoftwareBackendEnum.docker
            or backend == SoftwareBackendEnum.apptainer
        ):
            argv.append("--use-singularity")
        elif backend == SoftwareBackendEnum.envmodules:
            argv.append("--use-envmodules")
        elif backend == SoftwareBackendEnum.conda:
            argv.append("--use-conda")

        for key, value in snakemake_kwargs.items():
            if isinstance(value, bool):
                if value:  # Add flag only if True
                    argv.append(f"--{key}")
            else:
                argv.extend([f"--{key}", str(value)])

        return argv



import logging

from typing import Optional

from omnibenchmark.benchmark import Benchmark
from .RemoteStorage import StorageOptions

try:
    from .MinIOStorage import MinIOStorage
    from .S3config import S3_access_config_from_env

    S3_AVAILABLE = True
except ImportError:
    logger = logging.getLogger(__name__)
    logger.warning(
        "S3 storage not available. You might want to install the 'minio' and 'boto3' packages."
    )
    S3_AVAILABLE = False


# XXX revisit this, conceptually. Here we're mixing the storage API with the concrete
# MinIO implementation. We should use a factory pattern to create the appropriate storage object instead,
# assuming we support multiple storage types.
def get_storage(
    storage_type: str, auth_options: dict, benchmark: str, storage_options: StorageOptions = StorageOptions(out_dir="out")
) -> Optional["MinIOStorage"]:
    """
    Selects a remote storage type.

    Args:
    - storage_type (str): The type of the remote storage.
    - auth_options (dict): The authentication options.
    - benchmark (str): The benchmark name.

    Returns:
    - Optional[MinIOStorage]: The remote storage object, or None if unavailable.
    """
    if storage_type.upper() == "MINIO" or storage_type.upper() == "S3":
        return MinIOStorage(auth_options, benchmark, storage_options)


def get_storage_from_benchmark(benchmark: Benchmark) -> "MinIOStorage":
    """
    Selects a remote storage type from a benchmark object.

    Args:
    - benchmark (Benchmark): The benchmark object.

    Returns:
    - RemoteStorage: The remote storage object.
    """
    auth_options = remote_storage_args(benchmark)
    # setup storage
    return get_storage(
        str(benchmark.converter.model.storage_api),
        auth_options,
        str(benchmark.converter.model.storage_bucket_name),
        StorageOptions(out_dir="out")
    )


def remote_storage_args(benchmark: Benchmark) -> dict:
    if (
        str(benchmark.converter.model.storage_api).upper() == "MINIO"
        or str(benchmark.converter.model.storage_api).upper() == "S3"
    ):
        auth_options = S3_access_config_from_env()
        if benchmark.converter.model.storage is None:
            return {}
        base_dic = {
            "endpoint": benchmark.converter.model.storage,
            "secure": (
                True if benchmark.converter.model.storage.startswith("https") else False
            ),
        }
        if "access_key" in auth_options and "secret_key" in auth_options:
            auth_dic = {
                "access_key": auth_options["access_key"],
                "secret_key": auth_options["secret_key"],
            }
        else:
            auth_dic = {}

        return {**base_dic, **auth_dic}
    else:
        return {}


def remote_storage_snakemake_args(benchmark: Benchmark) -> dict:
    if (
        str(benchmark.converter.model.storage_api).upper() == "MINIO"
        or str(benchmark.converter.model.storage_api).upper() == "S3"
    ):
        auth_options = S3_access_config_from_env()
        return {
            "default-storage-provider": "s3",
            "default-storage-prefix": f"s3://{benchmark.converter.model.storage_bucket_name}",
            "storage-s3-endpoint-url": benchmark.converter.model.storage,
            "storage-s3-access-key": auth_options["access_key"],
            "storage-s3-secret-key": auth_options["secret_key"],
        }
    else:
        return {}



"""Base class for remote storage."""

from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Dict, Union

import packaging.version
from packaging.version import Version

from omnibenchmark.io.exception import RemoteStorageInvalidInputException


class StorageOptions:
    def __init__(self, out_dir: str):
        self.tracked_directories = [out_dir, "versions", "config", "software"]
        self.results_directories = [out_dir]
        self.extra_files_to_version_not_in_benchmark_yaml = [
            f"{out_dir}/**/parameters.json",
            f"{out_dir}/**/parameters_dict.tsv",
        ]  # glob style

        assert all(
            [
                isinstance(i, str) and i in self.tracked_directories
                for i in self.results_directories
            ]
        )


def is_valid_version(version: str):
    try:
        packaging.version.parse(version)
        return True
    except packaging.version.InvalidVersion:
        return False
    except TypeError:
        return False


class RemoteStorage(metaclass=ABCMeta):
    """
    A class representing a remote storage.

    Attributes:
    - version (str): The version of the current benchmark of the format: `major.minor` (for example: `0.1`).
    - versions (list): A list of versions.
    - files (dict): A dictionary of files.
    - benchmark (str): The current benchmark.
    - auth_options (dict): The authentication options.
    - storage_options (dict): The storage options.

    Methods:
    - __init__(auth_options, benchmark): Initializes the RemoteStorage object.
    - _parse_benchmark(benchmark): Parses and validates the benchmark name.
    - _parse_auth_options(auth_options): Parses and validates the authentication options.
    - connect(): Connects to the remote storage.
    - _test_connect(): Tests the connection to the remote storage.
    - _get_containers(): Retrieves the available containers.
    - _get_benchmarks(): Retrieves the available benchmarks.
    - _create_benchmark(benchmark): Creates a new benchmark.
    - _get_versions(): Retrieves the available versions of the current benchmark.
    - _parse_version(version): Parses and validates the current version.
    - set_version(version): Sets the current version of the benchmark.
    - _create_new_version(): Creates a new version of the benchmark.
    - _get_objects(readonly): Retrieves the objects in the storage for the current benchmark.
    - archive_version(version): Archives a specific benchmark version.
    - delete_version(version): Deletes a specific benchmark version.
    """

    def __init__(
        self,
        auth_options: Dict,
        benchmark: str,
        storage_options: StorageOptions,
    ):
        self.version = None
        self.versions = list()
        self.files = dict()
        self._parse_benchmark(benchmark)
        self._parse_auth_options(auth_options)
        self._parse_storage_options(storage_options)

    def _parse_benchmark(self, benchmark: str) -> None:
        if not isinstance(benchmark, str):
            # XXX this is what strict type checking is for :)
            raise RemoteStorageInvalidInputException("benchmark must be a string")
        self.benchmark = benchmark

    def _parse_auth_options(self, auth_options: Dict) -> None:
        if not isinstance(auth_options, dict):
            raise RemoteStorageInvalidInputException(
                "auth_options must be a dictionary"
            )
        self.auth_options = auth_options

    def _parse_storage_options(self, storage_options: StorageOptions) -> None:
        expected_attributes = [
            t for t in dir(storage_options) if not t.startswith("_")
        ]
        if not all([hasattr(storage_options, t) for t in expected_attributes]):
            raise RemoteStorageInvalidInputException(
                "storage_options must be a StorageOptions object"
            )
        self.storage_options = storage_options

    @abstractmethod
    def connect(self):
        """
        Connects to the storage Service.

        Returns:
            - Storage Client / Service
        """
        NotImplementedError

    @abstractmethod
    def _test_connect(self) -> None:
        """
        This method is used to test the connection to the remote storage.
        It does not perform any actual operations, but checks if the connection is successful.
        """
        NotImplementedError

    @abstractmethod
    def _create_benchmark(self, benchmark: str, update: bool = True) -> None:
        """
        Creates a new benchmark in the remote storage.

        Args:
            benchmark: The name of the benchmark.
            update: Whether to update the list of benchmarks. Defaults to True.
        """
        NotImplementedError

    @abstractmethod
    def _get_versions(self, update: bool = True, readonly: bool = False) -> None:
        """
        Retrieves the benchmark versions in the remote storage.

        Args:
            update (bool, optional): Whether to update the versions. Defaults to True.
            readonly (bool, optional): Whether to retrieve the versions in read-only mode. Defaults to False.
        """
        self.versions = []

    def _parse_version(self, version: Union[None, str] = None) -> Version:
        """
        Parses the current version based on the provided major and minor versions.

        Args:
            version (None or str, optional]): The version number. Defaults to None which returns the latest version.

        Returns:
            Version: The parsed version.

        Raises:
            RemoteStorageInvalidInputException: If an invalid version is provided.
        """
        if self.versions == []:
            self._get_versions()
        if version is None:
            if self.versions == []:
                version_out = Version("0.1")
            else:
                max_current_version = max(self.versions)
                version_out = Version(
                    f"{max_current_version.major}.{max_current_version.minor + 1}"
                )
        else:
            version_out = Version(version)
        return version_out

    def set_version(self, version: Union[None, str] = None) -> None:
        """
        Sets the current version of the remote storage.

        Args:
            version (None or str, optional): The version number as a string. Defaults to None which sets the latest version.

        Raises:
            RemoteStorageInvalidInputException: If version does not exist in the available versions.
        """
        self.version = self._parse_version(version)

    @abstractmethod
    def create_new_version(self):
        """
        Creates a new version of the benchmark.
        """
        NotImplementedError

    @abstractmethod
    def _get_objects(self):
        """
        Retrieves the objects in the storage for the current benchmark version.

        Args:
            readonly (bool, optional): Whether to retrieve the objects in read-only mode. Defaults to False.
        """
        NotImplementedError

    @abstractmethod
    def download_object(self, object_name: str, local_path: str):
        """
        Downloads an object from the storage.

        Args:
            object_name (str): The name of the object.
            local_path (str): The local path to download the object.
        """
        NotImplementedError

    @abstractmethod
    def archive_version(
        self,
        benchmark: str,
        outdir: Path = Path(),
        config: bool = True,
        code: bool = False,
        software: bool = False,
        results: bool = False,
    ):
        """
        Archives/Freezes a specific benchmark version.

        Args:
            version (str): The version to archive.
        """
        NotImplementedError

    @abstractmethod
    def delete_version(self, version):
        """
        Deletes a specific benchmark version.

        Args:
            version (str): The version to delete.
        """
        NotImplementedError



"""Configuration to set up a local cache and a datadir for test data download"""

# TODO: make sure this is in use

import configparser
import os
import platform
from typing import Optional, Any

from pathlib import Path

APP_NAME = "omnibenchmark"

_home = os.path.expanduser("~")

xdg_config_home = os.environ.get("XDG_CONFIG_HOME") or os.path.join(_home, ".config")
xdg_bench_home = os.environ.get("XDG_DATA_HOME") or os.path.join(
    _home, ".local", "share"
)


default_cfg = {"dirs": {"datasets": f"~/{APP_NAME}/datasets"}}

bench_dir = os.path.join(xdg_bench_home, APP_NAME)

if platform.system() == "Darwin":
    # macOS
    config_dir = Path("~/Library/Application Support/omnibenchmark").expanduser()
else:
    # Linux or others
    config_dir = Path(os.path.join(xdg_config_home, APP_NAME))


def get_config_file():
    return config_dir / f"{APP_NAME}.cfg"


def init_dirs():
    os.makedirs(bench_dir, exist_ok=True)
    os.makedirs(config_dir, exist_ok=True)


# def _get_config(config_path):
#    config = configparser.ConfigParser()
#    config.read(config_path)
#    return config


class ConfigAccessor:
    """
    A dict-like accessor for configuration files.

    This class provides a way to access configuration options with a dictionary-like
    interface while handling missing sections or keys gracefully.

    Usage:
        config = ConfigAccessor()
        value = config.get('section', 'key', fallback='default')
    """

    def __init__(self, config_path: Optional[Path] = None):
        """
        Initialize a ConfigAccessor with an optional config file path.

        Args:
            config_path: Path to the configuration file. If None, uses the default path.
        """
        if config_path is None:
            self.config_path = get_config_file()
        else:
            self.config_path = config_path

        # Ensure the config directory exists
        init_dirs()

        # Initialize the config parser
        self.config = configparser.ConfigParser()
        if self.config_path.exists():
            self.config.read(self.config_path)

    def get(self, section: str, key: str, default: Any = None) -> Any:
        """
        Get a configuration value from the specified section and key.

        Args:
            section: The configuration section
            key: The configuration key
            default: Value to return if the section or key doesn't exist

        Returns:
            The configuration value if it exists, otherwise the default value
        """
        try:
            return self.config[section][key]
        except (KeyError, configparser.NoSectionError, configparser.NoOptionError):
            return default

    def set(self, section: str, key: str, value: str) -> None:
        """
        Set a configuration value.

        Args:
            section: The configuration section
            key: The configuration key
            value: The value to set
        """
        if not self.config.has_section(section):
            self.config.add_section(section)

        self.config[section][key] = value

    def save(self) -> None:
        """
        Save the current configuration to the config file.
        """
        with open(self.config_path, "w") as configfile:
            self.config.write(configfile)

    def sections(self) -> list:
        """
        Get all available sections in the config.

        Returns:
            List of section names
        """
        return self.config.sections()

    def options(self, section: str) -> list:
        """
        Get all options (keys) in a section.

        Args:
            section: The section name

        Returns:
            List of options in the section or empty list if section doesn't exist
        """
        try:
            return self.config.options(section)
        except configparser.NoSectionError:
            return []


# Create a global config accessor instance
config = ConfigAccessor()



import sys

import click
import yaml

from pathlib import Path

from omnibenchmark.benchmark import Benchmark
from omnibenchmark.cli.utils.logging import logger


## to validate the YAML
def validate_benchmark(
    benchmark_file: str, out_dir: str, echo: bool = True
) -> Benchmark | None:
    if benchmark_file.endswith(".yaml") or benchmark_file.endswith(".yml"):
        try:
            with open(benchmark_file, "r") as file:
                yaml.safe_load(file)
                benchmark = Benchmark(Path(benchmark_file), Path(out_dir))

                if echo:
                    logger.info("Benchmark YAML file integrity check passed.")

                return benchmark

        except ValueError as e:
            logger.error(
                f"Error: Failed to parse YAML as a valid OmniBenchmark: {str(e)}.",
            )
            sys.exit(1)

        except yaml.YAMLError as e:
            logger.info(f"Error: YAML file format error: {e}.")
            click.Abort()

        except FileNotFoundError:
            logger.error(
                "Error: Benchmark YAML file not found.",
            )
            sys.exit(1)

        except Exception as e:
            logger.error(
                f"Error: An unexpected error occurred: {e}",
            )
            sys.exit(1)

    else:
        logger.error(
            "Error: Invalid benchmark input. Please provide a valid YAML file path.",
        )
        sys.exit(1)



from typing import Final

from humanfriendly import parse_timespan

# DEFAULT_TIMEOUT_HUMAN is the human representation of the max value we'll try to enforce
# as a timeout for the spawned task.
# This is currently only enforced on local environments; when supporting slurm we can pass
# it as runtime resource.
DEFAULT_TIMEOUT_HUMAN: Final[str] = "4h"
DEFAULT_TIMEOUT_SECONDS: Final[int] = int(parse_timespan(DEFAULT_TIMEOUT_HUMAN))

# LOCAL_TIMEOUT_VAR is the name of the environment variable that will be used to set the timeout
# for the spawned task on local environments. Snakemake will act weirdly if we reuse either 'runtime' or 'timeout',
# so don't be tempted to change that.
LOCAL_TIMEOUT_VAR: Final[str] = "local_task_timeout"



"""Sync benchmark definition file"""

import json
import os
import requests

from .config import bench_dir

base = "https://github.com/"
omni_essentials = "omnibenchmark/omni_essentials/-/raw/master/benchmarks/"
bench_definition = "benchmark_definition.yaml"


def get_bench_definition(bench_name: str, version: str, force: bool = False):
    bench_yaml = f"{bench_name}/{version}/{bench_definition}"
    local_bench_yaml = bench_dir + bench_yaml
    remote_bench_yaml = base + omni_essentials + f"{bench_yaml}?inline=false"
    if force or not os.path.isfile(local_bench_yaml):
        r = requests.get(remote_bench_yaml)
        data = r.json()
        with open(local_bench_yaml, "w") as f:
            json.dump(data, f)



import os
from pathlib import Path
from typing import Dict, List, Tuple, Union
import glob

from omnibenchmark.benchmark import Benchmark
from omnibenchmark.io.RemoteStorage import StorageOptions
from omnibenchmark.io.exception import MinIOStorageVersioningCorruptionException


def get_objects_to_tag(
    objdic: Dict,
    storage_options: StorageOptions,
) -> Tuple[List, List]:
    """
    Get a list of objects that need to be tagged with the current version.

    Args:
    - objdic: A dictionary of objects and associated metadata.
    - storage_options: The storage options.

    Returns:
    - A list of object names and a list of version ids.
    """
    object_names = sorted(list(objdic.keys()))
    object_names_to_tag = []
    versionid_of_objects_to_tag = []
    for object_name in object_names:
        # get newest version
        newest_version = sorted(
            objdic[object_name].items(),
            key=lambda it: it[1]["last_modified"],
            reverse=True,
        )[0][0]
        # get root directory
        object_name_red = object_name
        while not os.path.split(object_name_red)[0] == "":
            object_name_red = os.path.split(object_name_red)[0]
            if object_name_red in storage_options.tracked_directories:
                break
        # check if newest version exists and
        # if object is in tracked directories
        if (object_name_red in storage_options.tracked_directories) and (
            not objdic[object_name][newest_version]["is_delete_marker"]
        ):
            object_names_to_tag.append(object_name)
            versionid_of_objects_to_tag.append(newest_version)
    return object_names_to_tag, versionid_of_objects_to_tag


def get_expected_benchmark_output_files(
    benchmark: Benchmark,
    storage_options: StorageOptions,
) -> List:
    object_names_to_keep = benchmark.get_output_paths()
    if storage_options.extra_files_to_version_not_in_benchmark_yaml:
        for (
            glob_expression
        ) in storage_options.extra_files_to_version_not_in_benchmark_yaml:
            found_files = glob.glob(glob_expression, recursive=True)
            if isinstance(found_files, str):
                found_files = [found_files]
            for found_file in found_files:
                object_names_to_keep.add(found_file)
    return list(object_names_to_keep)


# filter objects based on workflow.
def filter_objects_to_tag(
    object_names_to_tag: List,
    versionid_of_objects_to_tag: List,
    storage_options: StorageOptions,
    benchmark: Union[None, Benchmark] = None,
) -> Tuple[List, List]:
    if benchmark is None:
        return object_names_to_tag, versionid_of_objects_to_tag
    else:
        object_names_to_keep = get_expected_benchmark_output_files(
            benchmark, storage_options
        )

        rootdirs = [Path(obj).parents[-2].name for obj in object_names_to_tag]
        overwrite_dirs_to_keep = [
            t
            for t in storage_options.tracked_directories
            if t not in storage_options.results_directories
        ]
        return zip(
            *[
                (obj, versionid_of_objects_to_tag[i])
                for i, obj in enumerate(object_names_to_tag)
                if obj in object_names_to_keep or rootdirs[i] in overwrite_dirs_to_keep
            ]
        )


def get_single_remoteversion_from_bmversion(
    di: Dict, object_name: str, query_version: str
) -> Union[str, None]:
    """
    Get the object version where the tag matches the benchmark version.

    Args:
    - di: A dictionary of objects and associated metadata.
    - object_name: The name of the object.
    - query_version: The benchmark version to query.

    Returns:
    - The object version or None where the tag matches the benchmark version.
    """
    version_ls = list(
        filter(
            lambda v: query_version in di[object_name][v]["tags"].keys(),
            di[object_name].keys(),
        )
    )
    if len(version_ls) > 1:
        raise MinIOStorageVersioningCorruptionException(
            f"Multiple versions found for object {object_name}"
        )
    return version_ls[0] if version_ls else None


def get_remoteversion_from_bmversion(di: Dict, query_version: str) -> List:
    """
    Get the versions of all objects where the tag matches the benchmark version.

    Args:
    - di: A dictionary of objects and associated metadata.
    - query_version: The benchmark version to query.

    Returns:
    - A list of objects and associated versions where the tag matches the benchmark
    """
    summary_ls = []
    for object_name in di.keys():
        version_id = get_single_remoteversion_from_bmversion(
            di, object_name, query_version
        )
        if version_id is not None:
            summary_ls.append(
                [
                    object_name,
                    version_id,
                    di[object_name][version_id]["last_modified"],
                    di[object_name][version_id]["size"],
                    di[object_name][version_id]["etag"],
                ]
            )
    return summary_ls


def prepare_csv_remoteversion_from_bmversion(summary_ls: List) -> str:
    """
    Prepare a CSV string from a list of objects and associated versions.
    """
    outstr = "name,version_id,last_modified,size,etag\n"
    for element in summary_ls:
        outstr += f"{element[0]},{element[1]},{element[2]},{element[3]},{element[4]}\n"
    return outstr



"""cli commands related to benchmark/module execution and start"""

import os
import sys
from itertools import chain
from pathlib import Path

import click
import humanfriendly

from omnibenchmark.benchmark.constants import DEFAULT_TIMEOUT_HUMAN
from omnibenchmark.cli.utils.logging import logger
from omnibenchmark.cli.utils.validation import validate_benchmark
from omnibenchmark.io.storage import get_storage_from_benchmark
from omnibenchmark.io.storage import remote_storage_snakemake_args
from omnibenchmark.workflow.snakemake import SnakemakeEngine
from omnibenchmark.workflow.workflow import WorkflowEngine

from .debug import add_debug_option


@click.group(name="run")
@click.pass_context
def run(ctx):
    """Run benchmarks or benchmark modules."""
    ctx.ensure_object(dict)


@add_debug_option
@run.command(name="benchmark")
@click.option(
    "-b",
    "--benchmark",
    help="Path to benchmark yaml file or benchmark id.",
    required=True,
    envvar="OB_BENCHMARK",
    type=click.Path(exists=True),
)
@click.option(
    "-c",
    "--cores",
    help="Use at most N CPU cores in parallel. Default is 1.",
    type=int,
    default=1,
)
@click.option(
    "-u",
    "--update",
    help="Force re-run execution for all modules and stages.",
    is_flag=True,
    default=False,
)
@click.option("-d", "--dry", help="Dry run.", is_flag=True, default=False)
@click.option(
    "-y",
    "--yes",
    help="Automatically confirm all prompts.",
    is_flag=True,
    default=False,
)
@click.option(
    "-k",
    "--continue-on-error",
    help="Go on with independent jobs if a job fails (--keep-going in snakemake).",
    is_flag=True,
    default=False,
)
@click.option(
    "-l",
    "--local",
    help="Execute and store results locally. Default False.",
    is_flag=True,
    default=False,
)
@click.option(
    "--keep-module-logs/--no-keep-module-logs",
    default=False,
    help="Keep module-specific log files after execution.",
)
@click.option(
    "--task-timeout",
    type=str,
    default=DEFAULT_TIMEOUT_HUMAN,
    help="Timeout for each separate task execution (local only). Do note that total runtime is not additive.",
)
@click.option(
    "--out-dir", type=str, default="out", help="Output folder name (local only)."
)
@click.pass_context
def run_benchmark(
    ctx,
    benchmark,
    cores,
    update,
    dry,
    yes,
    local,
    keep_module_logs,
    continue_on_error,
    task_timeout,
    out_dir,
):
    """Run a benchmark as specified in the yaml."""
    ctx.ensure_object(dict)

    # Retrieve the global debug flag from the Click context
    debug = ctx.obj.get("DEBUG", False)
    try:
        timeout_s = int(humanfriendly.parse_timespan(task_timeout))
    except humanfriendly.InvalidTimespan:
        logger.error(f"Invalid timeout value: {task_timeout}")
        sys.exit(1)

    # Validate out_dir usage
    if not local and out_dir != "out":
        logger.error("-Invalid arguments: --out-dir can only be used with --local")
        sys.exit(1)

    b = validate_benchmark(benchmark, out_dir)
    if b is None:
        # this should not happen, because validate raises, but that's not proper behavior.
        # We should sys.exit from here instead. But just to keep the signature valid.
        log_error_and_quit(logger, "Invalid benchmark file")
        return

    assert b is not None

    # it is as --continue_on_error originally
    if continue_on_error and not yes:
        msg = "run the full benchmark even if some jobs fail"
        abort_if_user_does_not_confirm(msg, logger)

    if update and not dry and not yes:
        msg = "re-run the entire workflow"
        abort_if_user_does_not_confirm(msg, logger)

    if not local:
        storage_options = remote_storage_snakemake_args(b)
        # creates bucket if it doesn't exist
        _ = get_storage_from_benchmark(b)
    else:
        storage_options = {}

    workflow: WorkflowEngine = SnakemakeEngine()

    # Controlling resource allocation with Snakemake is tricky
    # -c only controls the number of parallelism for the Snakemake scheduler,
    # but there's no built-in mechanism in local execution to enforce well-behaved resource
    # allocation. Adding to that, bioinformatics tools will not always allow to control cores/threads

    logger.info("Running benchmark...")
    success = workflow.run_workflow(
        b,
        cores=cores,
        update=update,
        dryrun=dry,
        continue_on_error=continue_on_error,
        keep_module_logs=keep_module_logs,
        backend=b.get_benchmark_software_backend(),
        debug=debug,
        out_dir=out_dir,
        local_timeout=timeout_s,
        **storage_options,
    )

    log_result_and_quit(logger, success, type="Benchmark")


@run.command(name="module")
@click.option(
    "-b",
    "--benchmark",
    help="Path to benchmark yaml file or benchmark id.",
    required=True,
    envvar="OB_BENCHMARK",
    type=click.Path(exists=True),
)
@click.option("-m", "--module", help="Module id to execute", type=str, required=True)
@click.option(
    "-i",
    "--input_dir",
    help="Path to the folder with the appropriate input files.",
    type=click.Path(exists=True, writable=True),
    default=None,
)
@click.option("-d", "--dry", help="Dry run.", is_flag=True, default=False)
@click.option(
    "-u",
    "--update",
    help="Force re-run execution for all modules and stages.",
    is_flag=True,
    default=False,
)
@click.option(
    "-k",
    "--continue-on-error",
    help="Go on with independent jobs if a job fails (--keep-going in snakemake).",
    is_flag=True,
    default=False,
)
@click.option(
    "--keep-module-logs/--no-keep-module-logs",
    default=False,
    help="Keep module-specific log files after execution.",
)
@click.pass_context
def run_module(
    ctx, benchmark, module, input_dir, dry, update, keep_module_logs, continue_on_error
):
    """
    Run a specific module that is part of the benchmark.
    """
    behaviours = {"input": input_dir, "example": None, "all": None}

    non_none_behaviours = {
        key: value for key, value in behaviours.items() if value is not None
    }
    if len(non_none_behaviours) >= 2:
        log_error_and_quit(
            logger,
            "Error: Only one of '--input_dir', '--example', or '--all' should be set. Please choose only one option.",
        )
        return

    # Construct a message specifying which option is set
    behaviour = list(non_none_behaviours)[0] if non_none_behaviours else None

    if behaviour == "example" or behaviour == "all":
        howmany = "all" if behaviour == "all" else "a"
        logger.info(f"Running module on {howmany} predefined remote example dataset.")

        # TODO Check how snakemake storage decorators work, do we have caching locally or just remote?
        # TODO Implement remote execution using remote url from benchmark definition
        log_error_and_quit(
            logger,
            "Error: Remote execution is not supported yet. Workflows can only be run in local mode.",
        )
        return

    logger.info("Running module on a local dataset.")

    b = validate_benchmark(benchmark, "/tmp")
    if b is None:
        # this should not happen, because validate raises, but that's not proper behavior. We should sys.exit
        # from here instead. But just to keep the signature valid.
        log_error_and_quit(logger, "Error: Invalid benchmark file.")
        return

    assert b is not None
    benchmark_nodes = b.get_nodes_by_module_id(module_id=module)

    is_entrypoint_module = all([node.is_entrypoint() for node in benchmark_nodes])

    if len(benchmark_nodes) <= 0:
        log_error_and_quit(
            logger,
            f"Error: Could not find module with id `{module}` in benchmark definition",
        )
        return

    assert len(benchmark_nodes) > 0
    logger.info(f"Found {len(benchmark_nodes)} workflow nodes for module {module}.")

    if not is_entrypoint_module and len(non_none_behaviours) == 0:
        log_error_and_quit(
            logger,
            "Error: At least one option must be specified. Use '--input_dir', '--example', or '--all'.",
        )
        return

    if input_dir is None:
        input_dir = os.getcwd()

    logger.info("Running module benchmark...")

    # Check if input path exists and is a directory
    if not (os.path.exists(input_dir) and os.path.isdir(input_dir)):
        log_error_and_quit(
            logger,
            f"Error: Input directory does not exist or is not a valid directory: `{input_dir}`",
        )
        return
    assert os.path.exists(input_dir) and os.path.isdir(input_dir)

    benchmark_datasets = b.get_benchmark_datasets()

    # Check available files in input to figure out what dataset are we processing
    if module in benchmark_datasets:
        # we're given the initial dataset module to process, then we know
        dataset = module
    else:
        # we try to figure the dataset based on the files present in the input directory
        files = os.listdir(input_dir)
        base_names = [file.split(".")[0] for file in files]
        dataset = next((d for d in benchmark_datasets if d in base_names), None)

    if dataset is None:
        log_error_and_quit(
            logger,
            f"Error: Could not infer the appropriate dataset to run the node workflow on based on the files available in `{input_dir}`. None of the available datasets {benchmark_datasets} match the base names of the files.",
        )
        return

    assert dataset is not None

    # Check if input directory contains all necessary input files
    required_inputs = list(map(lambda node: node.get_inputs(), benchmark_nodes))
    required_inputs = list(chain.from_iterable(required_inputs))
    required_input_files = list(
        set([os.path.basename(path) for path in required_inputs])
    )
    required_input_files = [
        file.format(dataset=dataset) for file in required_input_files
    ]

    input_files = os.listdir(input_dir)
    missing_files = [file for file in required_input_files if file not in input_files]

    if len(missing_files) > 0:
        log_error_and_quit(
            logger,
            f"Error: The following required input files are missing from the input directory: {missing_files}.",
        )
        return
    assert len(missing_files) == 0

    workflow: WorkflowEngine = SnakemakeEngine()

    for benchmark_node in benchmark_nodes:
        # When running a single module, it doesn't have sense to make parallelism level (cores) configurable
        success = workflow.run_node_workflow(
            node=benchmark_node,
            input_dir=Path(input_dir),
            dataset=dataset,
            cores=1,
            update=update,
            dryrun=dry,
            continue_on_error=continue_on_error,
            keep_module_logs=keep_module_logs,
            backend=b.get_benchmark_software_backend(),
        )
        log_result_and_quit(logger, success, type="Module")


@run.command(no_args_is_help=True, name="validate")
@click.option(
    "-b",
    "--benchmark",
    help="Path to benchmark yaml file or benchmark id.",
    envvar="OB_BENCHMARK",
    type=click.Path(exists=True),
)
@click.pass_context
def validate_yaml(ctx, benchmark):
    """Validate a benchmark yaml."""
    logger.info("Validating a benchmark yaml.")
    _ = validate_benchmark(benchmark, "/tmp")


def log_error_and_quit(logger, error):
    logger.error(error)
    sys.exit(1)


def log_result_and_quit(logger, success: bool, type: str):
    if success:
        logger.info(f"{type} run has finished successfully.")
        sys.exit(0)
    logger.error(f"{type} run has failed.")
    sys.exit(1)


def abort_if_user_does_not_confirm(msg: str, logger):
    _msg = f"Are you sure you want to {msg}?"
    if not click.confirm(_msg, abort=True):
        logger.debug("aborting")
        raise click.Abort()



import os.path
from pathlib import Path


class BenchmarkNode:
    def __init__(
        self,
        converter,
        stage,
        module,
        parameters,
        inputs,
        outputs,
        after=None,
    ):
        self.converter = converter
        self.stage = stage
        self.module = module
        self.parameters = parameters
        self.inputs = inputs
        self.outputs = outputs
        self.after = after

        self.stage_id = stage.id
        self.module_id = module.id
        self.param_id = (
            "default" if not self.parameters else f".{self.parameters.hash()}"
        )

    def is_entrypoint(self):
        return not self.inputs or len(self.inputs) == 0

    def get_id(self):
        return BenchmarkNode.to_id(
            self.stage_id, self.module_id, self.param_id, self.after
        )

    def get_benchmark_name(self):
        return self.converter.get_name()

    def get_benchmark_version(self):
        return self.converter.get_version()

    def get_benchmark_author(self):
        return self.converter.get_author()

    def get_benchmark_software_backend(self):
        return self.converter.get_software_backend()

    def get_benchmark_software_environments(self):
        return self.converter.get_software_environments()

    def get_definition(self):
        return self.converter.get_definition()

    def get_definition_file(self) -> Path:
        return self.converter.benchmark_file

    def get_inputs(self):
        return self.inputs.values() if self.inputs else []

    def get_inputs_dict(self):
        return self.inputs if self.inputs else {}

    def get_explicit_inputs(self):
        explicit_inputs = [
            self.converter.get_explicit_inputs(i)
            for i in self.converter.get_stage_implicit_inputs(self.stage)
        ]
        return explicit_inputs

    def get_input_paths(self, config, return_as_dict=False):
        input_paths = {}
        inputs = self.inputs.items() if self.inputs else {}
        for key, input in inputs:
            input = os.path.basename(input)
            input = input.replace("{dataset}", config["dataset"])
            input = os.path.join(config["input"], input)
            input_paths[key] = input

        if return_as_dict:
            return input_paths
        else:
            return list(input_paths.values())

    def get_outputs(self):
        return self.outputs if self.outputs else []

    def get_output_paths(self, config):
        output_paths = []

        pre = config.get("input", config.get("output"))
        dataset = config["dataset"]
        for output in self.get_outputs():
            output = output.format(
                pre=pre,
                dataset=dataset,
                stage=self.stage_id,
                module=self.module_id,
                params=self.param_id,
            )

            output_paths.append(output)

        return output_paths

    def get_benchmark_path(self, config):
        pre = config.get("input", config.get("output"))
        dataset = config["dataset"]

        output_paths = self.get_outputs()
        output_dir = os.path.commonpath(output_paths)
        if len(output_paths) == 1:
            output_dir = Path(os.path.dirname(output_dir))

        benchmark_path = Path(output_dir) / "{dataset}_performance.txt"

        benchmark_path = str(benchmark_path).format(
            pre=pre,
            dataset=dataset,
            stage=self.stage_id,
            module=self.module_id,
            params=self.param_id,
        )

        return benchmark_path

    def get_parameters(self):
        return self.parameters

    def get_repository(self):
        return self.converter.get_module_repository(module=self.module)

    def get_software_environment(self):
        return self.converter.get_module_environment(module=self.module)

    def is_initial(self):
        return self.converter.is_initial(self.stage)

    def get_stage(self):
        return self.stage

    def display_name(self):
        return BenchmarkNode.to_id(
            self.stage_id, self.module_id, self.param_id, self.after
        )

    def __str__(self):
        return self.display_name()

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        if isinstance(other, BenchmarkNode):
            return (self.stage_id, self.module_id, self.parameters, self.inputs) == (
                other.stage_id,
                other.module_id,
                other.parameters,
                other.inputs,
            )
        return False

    def __hash__(self):
        return hash(self.get_id())

    @staticmethod
    def to_id(stage_id, module_id, param_id, after_stage_id=None):
        node_id = f"{stage_id}-{module_id}-{param_id}"
        node_id += f"-after_{after_stage_id}" if after_stage_id else ""

        return node_id



import hashlib


# https://stackoverflow.com/a/3431838
def checksum(fname: str):
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()



class ValidationError(Exception):
    """Exception raised for validation errors."""

    def __init__(self, *errors):
        self.errors = errors[0] if len(errors) == 1 else errors
        super().__init__(*errors)

    def __str__(self):
        if isinstance(self.errors, str):
            return self.errors
        else:
            return "\n".join(map(str, self.errors))



import re
import os.path

from pathlib import Path
from typing import Dict, List, Set

from pydantic import ValidationError
import pydot

from omnibenchmark.benchmark import dag
from omnibenchmark.benchmark.converter import LinkMLConverter
from omnibenchmark.benchmark.validation.validator import Validator
from omnibenchmark.utils import format_mc_output


class Benchmark:
    def __init__(self, benchmark_yaml: Path, out_dir: Path = Path("out")):
        self.directory = benchmark_yaml.parent.absolute()

        if not os.path.exists(out_dir):
            os.mkdir(out_dir)

        converter = LinkMLConverter(benchmark_yaml)

        validator = Validator()
        try:
            converter = validator.validate(self.directory, converter)
        except ValidationError:
            # Just re-raising the validation error, explicitly
            raise

        self.converter: LinkMLConverter = converter
        self.out_dir = out_dir
        self.G = dag.build_benchmark_dag(converter, self.out_dir)

        self.execution_paths = None

    def get_converter(self):
        return self.converter

    def get_benchmark_name(self):
        return self.converter.get_name()

    def get_benchmark_version(self):
        return self.converter.get_version()

    def get_benchmark_author(self):
        return self.converter.get_author()

    def get_benchmark_software_backend(self):
        return self.converter.get_software_backend()

    def get_benchmark_software_environments(self):
        return self.converter.get_software_environments()

    def get_definition(self):
        return self.converter.get_definition()

    def get_definition_file(self) -> Path:
        return self.converter.benchmark_file

    def get_easyconfigs(self):
        return self.converter.get_easyconfigs()

    def get_conda_envs(self):
        return self.converter.get_conda_envs()

    def get_nodes(self):
        return list(self.G.nodes)

    def get_stage_ids(self):
        return self.converter.get_stages().keys()

    def get_node_by_id(self, node_id):
        for node in self.G.nodes:
            if node.get_id() == node_id:
                return node

        return None

    def get_nodes_by_module_id(self, module_id: str) -> List:
        nodes = []
        for node in self.G.nodes:
            if node.module_id == module_id:
                nodes.append(node)

        return nodes

    def get_nodes_by_stage_id(self, stage_id: str) -> List:
        nodes = []
        for node in self.G.nodes:
            if node.stage_id == stage_id:
                nodes.append(node)

        return nodes

    def get_benchmark_datasets(self) -> List[str]:
        datasets = []
        for _, stage in self.converter.get_stages().items():
            # There should be only one initial stage
            if self.converter.is_initial(stage):
                for module in stage.modules:
                    datasets.append(module.id)

                break

        return datasets

    def get_execution_paths(self):
        if self.execution_paths is None:
            self.execution_paths = self._generate_execution_paths()

        return self.execution_paths

    def get_output_paths(self) -> Set[str]:
        execution_paths = self.get_execution_paths()

        output_paths = [
            _normalize_dataset_path(output, self.out_dir)
            for path in execution_paths
            for output in self._construct_output_paths(prefix=self.out_dir, nodes=path)
        ]

        return set(output_paths)

    def get_metric_collector_output_paths(self):
        collectors = self.get_metric_collectors()
        output_paths = []

        for collector in collectors:
            for output in collector.outputs:
                output_paths.append(
                    format_mc_output(output, self.out_dir, collector.id)
                )

        return set(output_paths)

    def get_explicit_inputs(self, stage_id: str):
        stage = self.converter.get_stage(stage_id)
        implicit_inputs = self.converter.get_stage_implicit_inputs(stage)
        explicit_inputs = [
            self.converter.get_explicit_inputs(i) for i in implicit_inputs
        ]
        return explicit_inputs

    def get_explicit_input(self, input_ids: List[str]) -> Dict[str, str]:
        return self.converter.get_explicit_inputs(input_ids)

    def get_explicit_outputs(self, stage_id: str):
        stage = self.converter.get_stage(stage_id)
        return self.converter.get_stage_outputs(stage)

    def get_available_parameter(self, module_id: str):
        node = next(node for node in self.G.nodes if node.module_id == module_id)
        return node.get_parameters()

    def get_metric_collectors(self):
        return self.converter.get_metric_collectors()

    def export_to_dot(self) -> pydot.Dot:
        pydot_graph = dag.export_to_dot(
            self.G,
            title=self.get_benchmark_name(),
        )

        return pydot_graph

    def export_to_mermaid(self, show_params: bool = True) -> str:
        # Initialize the mermaid diagram syntax
        mermaid_diagram = "---\n"
        mermaid_diagram += f"title: {self.get_benchmark_name()}\n"
        mermaid_diagram += "---\n"
        mermaid_diagram += "flowchart LR\n"

        # Graph customisation
        mermaid_diagram += "\tclassDef param fill:#f96\n"

        module_id_params_dict = {}
        for stage_id, stage in self.converter.get_stages().items():
            mermaid_diagram += f"\tsubgraph {stage_id}\n"

            for module_id, module in self.converter.get_modules_by_stage(stage).items():
                module_parameters = self.converter.get_module_parameters(module)
                if module_parameters:
                    module_id_params_dict[module_id] = module_parameters

                mermaid_diagram += f"\t\t{module_id}\n"
                module_nodes = self.get_nodes_by_module_id(module_id)
                from_nodes = [list(self.G.predecessors(node)) for node in module_nodes]
                from_node_module_ids = sorted(
                    list(
                        set(
                            [
                                node.module_id
                                for flatten in from_nodes
                                for node in flatten
                            ]
                        )
                    )
                )
                for from_module_id in from_node_module_ids:
                    mermaid_diagram += f"\t\t{from_module_id} --> {module_id}\n"
            mermaid_diagram += "\tend\n"

        if show_params:
            for module_id, params in module_id_params_dict.items():
                mermaid_diagram += f"\tsubgraph params_{module_id}\n"
                for param in params:
                    mermaid_diagram += f"\t\t{param.hash()}{param.to_cli_args()}\n"

                mermaid_diagram += "\tend\n"
                mermaid_diagram += f"\tparams_{module_id}:::param --o {module_id}\n"

        return mermaid_diagram

    def __str__(self):
        return f"Benchmark({self.get_definition})"

    def _generate_execution_paths(self):
        path_exclusions = self._get_path_exclusions()
        initial_nodes, terminal_nodes = dag.find_initial_and_terminal_nodes(self.G)

        execution_paths = []
        for initial_node in initial_nodes:
            for terminal_node in terminal_nodes:
                paths = dag.list_all_paths(self.G, initial_node, terminal_node)
                paths_after_exclusion = dag.exclude_paths(paths, path_exclusions)

                execution_paths.extend(paths_after_exclusion)

        return execution_paths

    def _get_path_exclusions(self):
        path_exclusions = {}
        stages = self.converter.get_stages()
        for stage_id in stages:
            stage = stages[stage_id]

            modules_in_stage = self.converter.get_modules_by_stage(stage)
            for module_id in modules_in_stage:
                module_excludes = self.converter.get_module_excludes(module_id)
                if module_excludes:
                    path_exclusions[module_id] = module_excludes

        return path_exclusions

    def _construct_output_paths(self, prefix, nodes):
        if nodes is None or len(nodes) == 0:
            # TODO: assert len >= 2
            return []
        else:
            head = nodes[0]
            tail = nodes[1:]
            stage_outputs = self.converter.get_stage_outputs(head.stage_id).values()

            current_path = f"{head.stage_id}/{head.module_id}"
            if any(["{params}" in o for o in stage_outputs]):
                current_path += f"/{head.param_id}"

            new_prefix = f"{prefix}/{current_path}"
            paths = [
                x.format(
                    input=prefix,
                    stage=head.stage_id,
                    module=head.module_id,
                    params=head.param_id,
                    dataset="{dataset}",
                )
                for x in stage_outputs
            ]

            return paths + self._construct_output_paths(new_prefix, tail)


def _normalize_dataset_path(path: str, prefix: Path) -> str:
    """Normalizes a benchmark output path by extracting and formatting the dataset identifier.

    Takes a path string containing a dataset identifier segment and formats it consistently.
    The path is expected to follow the structure: {prefix}/.../{dataset}/...

    Args:
        path: The output path string containing a dataset identifier
        prefix: The base directory Path object where outputs are stored

    Returns:
        A formatted path string with the dataset identifier properly substituted.
        Returns empty string if the path doesn't match the expected structure.

    Example:
        >>> normalize_dataset_path("out/stage1/dataset123/results", Path("out"))
        'out/stage1/dataset123/results'
    """
    pattern = rf"{prefix.as_posix()}/.+?/([^/]+)/.+?$"
    matched = re.match(pattern, path)
    if matched is None:
        # we probably should raise a ValueError here, but
        # this is better than an IndexError
        return ""
    dataset = matched[1]
    return path.format(dataset=dataset)



from typing import List, Optional

import yaml
from pydantic import BaseModel, HttpUrl
import spdx_license_list


class DerivedSoftware(BaseModel):
    """Model for sources that this work derives from"""

    description: str
    author: str
    url: HttpUrl
    doi: Optional[str] = None

    # TODO: should enforce URI-like, either file:// or https://
    easyconfig: Optional[str] = None
    conda: Optional[str] = None


class ModuleMetadata(BaseModel):
    """Metadata model for an Omnibenchmark Module"""

    name: str
    description: str
    author: str
    derives_from: List[DerivedSoftware]
    license: str

    def validate_license(self):
        """Validate that license is a recognized open source license"""
        if self.license not in spdx_license_list.LICENSES.keys():
            raise ValueError(
                f"License {self.license} not in recognized open source licenses"
            )

    @classmethod
    def from_yaml(cls, yaml_str: str):
        """Alternative constructor that loads from YAML string"""
        data = yaml.safe_load(yaml_str)
        return cls(**data)



class RemoteStorageException(Exception):
    """
    General Exception for RemoteStorage.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class RemoteStorageInvalidInputException(RemoteStorageException):
    """
    Exception raised for invalid inputs in RemoteStorage and descendants.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class MinIOStorageException(RemoteStorageException):
    """
    General Exception for MinIOStorage.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class MinIOStorageConnectionException(MinIOStorageException):
    """
    Exception raised for connection issues with S3.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class MinIOStorageBucketManipulationException(MinIOStorageException):
    """
    Exception raised for errors with bucket creation and general bucket manipulations.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class MinIOStorageVersioningCorruptionException(MinIOStorageException):
    """
    Exception raised for errors with object versioning.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)



import zipfile
from pathlib import Path
from typing import List
import os

from omni_schema.datamodel import omni_schema

from omnibenchmark.io.RemoteStorage import StorageOptions
from omnibenchmark.io.code import clone_module
from omnibenchmark.benchmark import Benchmark
from omnibenchmark.io.files import download_files, list_files


def prepare_archive_code(benchmark: Benchmark) -> List[Path]:
    """
    Prepare the code to archive and return list of all filenames.

    Args:
        benchmark (str): The benchmark id.

    Returns:
        List[Path]: The filenames of all code to archive.
    """

    nodes = benchmark.get_nodes()
    repositories = set()
    for node in nodes:
        repositories.add((node.get_repository().url, node.get_repository().commit))

    repositories_dir = Path(".snakemake") / "repos"
    files = []
    while repositories:
        repo = repositories.pop()
        files += list(clone_module(repositories_dir, repo[0], repo[1]).iterdir())
    return files


def prepare_archive_software(benchmark: Benchmark) -> List[Path]:
    """
    Prepare the software to archive and return list of all filenames.

    Args:
        benchmark (str): The benchmark id.

    Returns:
        List[Path]: The filenames of all software to archive.
    """
    # decide on software type
    which_env = benchmark.get_benchmark_software_backend()
    files = []
    if which_env != omni_schema.SoftwareBackendEnum.host:
        if which_env == omni_schema.SoftwareBackendEnum.envmodules:
            files += prepare_archive_software_easyconfig(benchmark)
        elif which_env == omni_schema.SoftwareBackendEnum.apptainer:
            files += prepare_archive_software_apptainer(benchmark)
        elif which_env == omni_schema.SoftwareBackendEnum.conda:
            files += prepare_archive_software_conda(benchmark)
        else:
            raise NotImplementedError(
                f"Software backend {which_env} not implemented yet."
            )
    return files


def prepare_archive_software_easyconfig(benchmark: Benchmark) -> List[Path]:
    """
    Prepare the software easyconfig to archive and return list of all filenames.

    Args:
        benchmark (str): The benchmark id.

    Returns:
        List[Path]: The filenames of all software easyconfig to archive.
    """
    files = []
    softenvs = benchmark.get_benchmark_software_environments()
    for softenv in softenvs.values():
        envmodule_file = (benchmark.directory / Path(softenv.envmodule)).relative_to(
            Path(os.getcwd())
        )
        if envmodule_file.is_file():
            files.append(envmodule_file)
        else:
            raise FileNotFoundError(f"File {envmodule_file} not found.")
        easyconfig_file = (
            benchmark.directory / Path(softenv.easyconfig_file)
        ).relative_to(Path(os.getcwd()))
        if easyconfig_file.is_file():
            files.append(easyconfig_file)
        else:
            raise FileNotFoundError(f"File {easyconfig_file} not found.")
    return files


def prepare_archive_software_conda(benchmark: Benchmark) -> List[Path]:
    """
    Prepare the conda to archive and return list of all filenames.

    Args:
        benchmark (str): The benchmark id.

    Returns:
        List[Path]: The filenames of all conda to archive.
    """
    # prepare conda, as in snakemake archive (https://github.com/snakemake/snakemake/blob/76d53290a003891c5ee41f81e8eb4821c406255d/snakemake/deployment/conda.py#L316) maybe?
    # return all files
    # from omnibenchmark.software.conda_backend import pin_conda_envs
    # pin_conda_envs(benchmark.get_definition_file())
    files = []
    softenvs = benchmark.get_benchmark_software_environments()
    for softenv in softenvs.values():
        conda_file = (benchmark.directory / Path(softenv.conda)).relative_to(
            Path(os.getcwd())
        )
        if conda_file.is_file():
            files.append(conda_file)
        else:
            raise FileNotFoundError(f"File {conda_file} not found.")
    return files


def prepare_archive_software_apptainer(benchmark: Benchmark) -> List[Path]:
    """
    Prepare the software apptainer to archive and return list of all filenames.

    Args:
        benchmark (str): The benchmark id.

    Returns:
        List[Path]: The filenames of all software apptainer to archive.
    """
    # prepare apptainer, check if .sif file exists
    # return all files
    files = []
    softenvs = benchmark.get_benchmark_software_environments()
    for softenv in softenvs.values():
        apptainer_file = (benchmark.directory / Path(softenv.apptainer)).relative_to(
            Path(os.getcwd())
        )
        if apptainer_file.is_file():
            files.append(apptainer_file)
        else:
            raise FileNotFoundError(f"File {apptainer_file} not found.")
    return files


def prepare_archive_results(
    benchmark: Benchmark, results_dir: str, local: bool = False
) -> List[Path]:
    """
    Prepare the results to archive and return list of all filenames.

    Args:
        benchmark (str): The benchmark id.

    Returns:
        List[Path]: The filenames of all results to archive.
    """
    # get all results, check if exist locally, otherwise download

    objectnames, etags = list_files(
        benchmark.get_definition_file().as_posix(),
        type="all",
        stage="",
        module="",
        file_id="",
        local=local,
        storage_options=StorageOptions(out_dir=results_dir),
    )
    if not local:
        download_files(
            benchmark.get_definition_file().as_posix(),
            type="all",
            stage="",
            module="",
            file_id="",
            overwrite=True,
        )
    return objectnames


def archive_version(
    benchmark: Benchmark,
    outdir: Path = Path(),
    config: bool = True,
    code: bool = False,
    software: bool = False,
    results: bool = False,
    results_dir: str = "out",
    compression=zipfile.ZIP_STORED,
    compresslevel: int = None,
    dry_run: bool = False,
    local: bool = False,
):
    # retrieve all filenames to save
    filenames = []

    ## config (benchmark.yaml)
    if config:
        filenames.append(benchmark.get_definition_file())

    ## code (code files)
    ### check local cache of GH repos
    ### download repos if not in cache
    ### iterate over all files in repos
    if code:
        filenames += prepare_archive_code(benchmark)

    ## software (software files)
    ### easyconfig
    #### save easyconfig files
    ### conda
    #### save explicit packages list
    ### apptainer
    #### save .sif file
    if software:
        filenames += prepare_archive_software(benchmark)

    ## results (results files)
    ### check if results match remote, if not download
    if results:
        filenames += prepare_archive_results(benchmark, results_dir, local)

    if dry_run:
        return filenames
    else:
        match compression:
            case zipfile.ZIP_BZIP2:
                file_extension = ".bz2"
            case zipfile.ZIP_LZMA:
                file_extension = ".xz"
            case _:
                file_extension = ".zip"
        # save all files to zip archive
        outfile = f"{benchmark.get_benchmark_name()}_{benchmark.get_converter().get_version()}{file_extension}"
        with zipfile.ZipFile(
            outdir / outfile, "w", compression=compression, compresslevel=compresslevel
        ) as archive:
            for filename in filenames:
                archive.write(filename, filename)

        return outdir / outfile


# archive_version(benchmark, outdir=Path(), config=True, code=True, software=False, results=False)



from pathlib import Path

from snakedeploy import conda as snakeconda

"""
Conda-based software management, per benchmark yaml (and not per omniblock)
"""


def pin_conda_envs(benchmark_yaml: Path) -> None:
    """
    Pins a conda env file using snakedeploy
    """
    snakeconda.pin_conda_envs(
        conda_env_paths=[benchmark_yaml], conda_frontend="conda", warn_on_error=True
    )



# from: https://stackoverflow.com/a/1094933
# TODO(ben): use humanize
def sizeof_fmt(num: float, suffix: str = "B"):
    if abs(num) < 1024:
        return f"{num: 5d}{suffix}"
    for unit in ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"):
        if abs(num) < 1024:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024
    return f"{num:.1f}Yi{suffix}"



from functools import wraps

import click

from .utils.logging import configure_logging


def add_debug_option(cmd):
    """Decorator to add debug option to commands and groups"""
    if isinstance(cmd, click.Command):
        # For existing commands/groups
        if not any(param.name == "debug" for param in cmd.params):
            cmd.params.insert(
                0,
                click.Option(
                    ["--debug/--no-debug"],
                    is_eager=True,
                    expose_value=False,
                    callback=lambda ctx, param, value: _set_debug(ctx, value),
                    help="Enable debug mode",
                ),
            )
        return cmd

    # For functions that will become commands/groups
    @click.option(
        "--debug/--no-debug",
        is_eager=True,
        expose_value=False,
        callback=lambda ctx, param, value: _set_debug(ctx, value),
        help="Enable debug mode",
    )
    @wraps(cmd)
    def wrapper(*args, **kwargs):
        return cmd(*args, **kwargs)

    return wrapper


def _set_debug(ctx, value: bool):
    """Callback function for debug flag"""
    # Get the root context
    root_ctx = ctx.find_root()
    root_ctx.ensure_object(dict)

    # Get command path depth
    cmd_depth = len(ctx.command_path.split())

    # Initialize DEBUG if not set
    if "DEBUG" not in root_ctx.obj:
        root_ctx.obj["DEBUG"] = False

    # Allow setting debug to True from any level
    # Allow setting debug to False only from the level that set it to True
    if value is True or cmd_depth == 1:
        root_ctx.obj["DEBUG"] = value

    configure_logging(root_ctx.obj["DEBUG"])
    return root_ctx.obj["DEBUG"]



from itertools import groupby
from typing import Dict

import minio

from omnibenchmark.io.exception import MinIOStorageBucketManipulationException
from omnibenchmark.io.RemoteStorage import is_valid_version


def get_s3_object_versions_and_tags(
    client: minio.Minio, benchmark: str, readonly: bool = False
) -> Dict:
    """
    Retrieve the metadata of all objects in a S3 Bucket.

    Args:
    - client: A MinIO client object.
    - benchmark: The name of the S3 bucket.
    - readonly: A boolean indicating whether the client is read-only. Object tags can only be retrieved if the client is not read-only.

    Returns:
    - A dictionary of objects and associated metadata.
    """
    if not client.bucket_exists(benchmark):
        raise MinIOStorageBucketManipulationException(
            f"Benchmark {benchmark} does not exist."
        )

    object_ls = list(
        client.list_objects(
            benchmark, include_version=True, include_user_meta=True, recursive=True
        )
    )

    # Group listed objects by their name
    grouped_objects = groupby(object_ls, key=lambda o: o.object_name)

    di = {}

    # Iterate over object groups
    for object_name, objects in grouped_objects:
        objects = list(objects)
        di[object_name] = {}

        # Extract attributes for the objects group
        version_ids = [o.version_id for o in objects]
        sizes = [o.size for o in objects]
        last_modified = [o.last_modified for o in objects]
        is_delete_markers = [o.is_delete_marker for o in objects]
        etags = [o.etag for o in objects]

        # Iterate to all versions and create a dictionary entry
        for i, v in enumerate(version_ids):
            tags_filt = {}
            if not readonly and not is_delete_markers[i]:
                tags = client.get_object_tags(benchmark, object_name, version_id=v)
                if tags is not None:
                    tags_filt = {k: w for k, w in tags.items() if is_valid_version(k)}

            di[object_name][v] = {
                "tags": tags_filt,
                "size": sizes[i],
                "last_modified": last_modified[i],
                "is_delete_marker": is_delete_markers[i],
                "etag": etags[i],
            }
    return di



"""cli commands related to software management"""

import os
import sys
from pathlib import Path

import click
import yaml

from omnibenchmark.cli.utils.logging import logger


@click.group(name="software")
@click.pass_context
def software(ctx):
    """Manage and install benchmark-specific software."""
    ctx.ensure_object(dict)


## singularity #####################################################################################


@click.group
@click.pass_context
def singularity(ctx):
    """Manage and install software using Singularity."""
    ctx.ensure_object(dict)


software.add_command(singularity)


@singularity.command(name="build", no_args_is_help=True)
@click.option(
    "-e",
    "--easyconfig",
    help="Easyconfig.",
    required=True,
)
def singularity_build(easyconfig):
    """Build a singularity (fakeroot) image for a given easyconfig."""
    logger.info(
        f"Installing software for {easyconfig} within a Singularity container. It will take some time."
    )
    from omnibenchmark.software import common
    from omnibenchmark.software import easybuild_backend as eb

    if common.check_easybuild_status().returncode != 0:
        raise RuntimeError("ERROR: Easybuild not installed")
    if common.check_singularity_status().returncode != 0:
        raise RuntimeError("ERROR: Singularity not installed")

    ## check the easyconfig exists
    try:
        eb.get_easyconfig_full_path(easyconfig=easyconfig)
    except Exception:
        logger.error("ERROR: easyconfig not found.")
        sys.exit()

    ## do
    singularity_recipe = "Singularity_" + easyconfig + ".txt"
    envmodule_name = eb.get_envmodule_name_from_easyconfig(easyconfig)
    eb.create_definition_file(
        easyconfig=easyconfig,
        singularity_recipe=singularity_recipe,
        envmodule=envmodule_name,
        nthreads=len(os.sched_getaffinity(0)),
    )

    logger.info(
        "DONE: singularity recipe written for "
        + singularity_recipe
        + "\nDOING: building the image"
    )

    sb = eb.singularity_build(
        singularity_recipe=singularity_recipe, easyconfig=easyconfig
    )
    if sb.returncode != 0:
        logger.error("ERROR: " + sb.stderr)
    if sb.returncode == 0:
        logger.info(sb.stdout)
        logger.info("DONE: singularity image built for " + singularity_recipe)


@singularity.command(name="prepare")
@click.option(
    "-b",
    "--benchmark",
    help="Benchmark YAML.",
    type=click.Path(exists=True),
    required=True,
    envvar="OB_BENCHMARK",
)
def singularity_prepare(benchmark):
    """Build all singularity (fakeroot) images needed for a benchmark."""
    logger.info(
        f"Installing software for {benchmark} using Singularity containers. It will take some time."
    )
    from omnibenchmark.benchmark import Benchmark
    from omnibenchmark.software import common
    from omnibenchmark.software import easybuild_backend as eb

    if common.check_easybuild_status().returncode != 0:
        raise RuntimeError("ERROR: Easybuild not installed")
    if common.check_singularity_status().returncode != 0:
        raise RuntimeError("ERROR: Singularity not installed")

    with open(benchmark, "r") as fh:
        yaml.safe_load(fh)
        benchmark = Benchmark(Path(benchmark))

    for easyconfig in benchmark.get_easyconfigs():
        ## check the easyconfig exists
        try:
            eb.get_easyconfig_full_path(easyconfig=easyconfig)
        except Exception:
            logger.error("ERROR: easyconfig not found.")
            sys.exit()

        ## do
        singularity_recipe = "Singularity_" + easyconfig + ".txt"
        envmodule_name = eb.get_envmodule_name_from_easyconfig(easyconfig)
        eb.create_definition_file(
            easyconfig=easyconfig,
            singularity_recipe=singularity_recipe,
            envmodule=envmodule_name,
            nthreads=len(os.sched_getaffinity(0)),
        )

        logger.info(
            "DONE: singularity recipe written for "
            + singularity_recipe
            + "\nDOING: building the image"
        )

        sb = eb.singularity_build(
            singularity_recipe=singularity_recipe, easyconfig=easyconfig
        )
        if sb.returncode != 0:
            logger.error("ERROR: " + sb.stderr)
        if sb.returncode == 0:
            logger.info(sb.stdout)
            logger.info("DONE: singularity image built for " + singularity_recipe)
        logger.info("DONE: singularity images built.")


@singularity.command(name="push", no_args_is_help=True)
@click.option(
    "-u", "--docker_username", help="Docker username.", type=str, required=True
)
@click.option(
    "-p", "--docker_password", help="Docker password.", type=str, required=True
)
@click.option(
    "-s",
    "--sif",
    help="Path to the Singularity SIF file.",
    type=click.Path(writable=True),
    required=True,
)
@click.option(
    "-o",
    "--oras",
    help="Registry's ORAS static URL, for instance oras://registry.mygitlab.ch/myuser/myproject:mytag.",
    type=str,
    required=True,
)
def singularity_push(docker_username, docker_password, sif, oras):
    """Pushes a singularity SIF file to an ORAS-compatible registry."""
    logger.info(f"Pushing {sif} to the registry {oras}.")

    from omnibenchmark.software import easybuild_backend as eb

    # FIXME, this does not exist
    eb.push_to_registry(
        sif=sif,
        docker_username=docker_username,
        docker_password=docker_password,
        oras=oras,
    )
    logger.info("DONE\n.")


## envmodules ########################################################################################


@click.group
@click.pass_context
def module(ctx):
    """Manage and install software using Easybuild"""
    ctx.ensure_object(dict)


software.add_command(module)


@module.command(name="build", no_args_is_help=True)
@click.option(
    "-e",
    "--easyconfig",
    help="Easyconfig.",
    required=True,
)
@click.option("-p", "--threads", type=int, default=2, help="Number of threads.")
def envmodules_build(easyconfig, threads):
    """Build a given easyconfig (and generates the relevant envmodules)."""
    logger.info(
        f"\nInstalling software for {easyconfig} using easybuild. It will take some time."
    )

    from omnibenchmark.software import common
    from omnibenchmark.software import easybuild_backend as eb

    if common.check_easybuild_status().returncode != 0:
        raise RuntimeError("ERROR: Easybuild not installed")

    ## check the easyconfig exists
    try:
        eb.get_easyconfig_full_path(easyconfig=easyconfig)
    except Exception:
        print("ERROR: easyconfig not found.\n")
        sys.exit()

    p = eb.easybuild_easyconfig(easyconfig=easyconfig, threads=threads)
    if p.returncode != 0:
        logger.error("ERROR: " + p.stderr)
    if p.returncode == 0:
        logger.info(p.stdout)
    logger.info("DONE: built " + easyconfig)


@module.command(name="prepare", no_args_is_help=True)
@click.option(
    "-b",
    "--benchmark",
    help="Benchmark YAML.",
    type=click.Path(exists=True),
    required=True,
    envvar="OB_BENCHMARK",
)
@click.option("-p", "--threads", default=2, help="Number of threads.", type=int)
def envmodules_prepare(benchmark, threads):
    """Build all envmodules needed for a given benchmark YAML."""
    logger.info(
        f"Installing software for {benchmark} using envmodules. It will take some time."
    )
    from omnibenchmark.benchmark import Benchmark
    from omnibenchmark.software import common
    from omnibenchmark.software import easybuild_backend as eb

    if common.check_easybuild_status().returncode != 0:
        raise RuntimeError("ERROR: Easybuild not installed")
    if common.check_lmod_status().returncode != 0:
        raise RuntimeError("ERROR: lmod not installed")

    with open(benchmark, "r") as fh:
        yaml.safe_load(fh)
        benchmark = Benchmark(Path(benchmark))

    for easyconfig in benchmark.get_easyconfigs():
        logger.info(f"Building {easyconfig}. It will take some time.")
        ## check the easyconfig exists
        try:
            fp = eb.get_easyconfig_full_path(easyconfig=easyconfig)
        except Exception:
            logger.error(
                "ERROR: easyconfig "
                + fp
                + " not found, is it within your robot search path?",
            )
            sys.exit()

        p = eb.easybuild_easyconfig(easyconfig=easyconfig, threads=threads)
        if p.returncode != 0:
            logger.error("ERROR: " + p.stderr)
        if p.returncode == 0:
            logger.info(p.stdout)
            logger.info("DONE: built " + easyconfig)


## conda #############################################################################################


@click.group
@click.pass_context
def conda(ctx):
    """Manage and install software using conda."""
    ctx.ensure_object(dict)


software.add_command(conda)


@conda.command(name="pin")
@click.option(
    "-e",
    "--env",
    "conda_env",
    help="Conda env YAML.",
    type=click.Path(writable=True),
    required=True,
)
def pin_conda_env(conda_env):
    """Pin all conda env-related dependencies versions using snakedeploy."""
    logger.info(f"Pinning {conda_env} via snakedeploy. It will take some time.")
    from omnibenchmark.software import conda_backend

    conda_backend.pin_conda_envs(conda_env)
    logger.info(f"\nDONE: Pinned {conda_env}\n")


@conda.command(name="prepare")
@click.option(
    "-b",
    "--benchmark",
    help="Benchmark YAML.",
    type=click.Path(exists=True),
    required=True,
    envvar="OB_BENCHMARK",
)
def conda_prepare(benchmark):
    """Pin all conda envs needed for a given benchmark YAML."""
    logger.info(f"Pinning conda envs for {benchmark}. It will take some time.")
    from omnibenchmark.benchmark import Benchmark
    from omnibenchmark.software import common, conda_backend

    if common.check_conda_status().returncode != 0:
        raise RuntimeError("ERROR: conda not installed")

    with open(benchmark, "r") as fh:
        yaml.safe_load(fh)
        bm = Benchmark(Path(benchmark))

    for conda in bm.get_conda_envs():
        conda_yaml_path = os.path.join(os.path.dirname(benchmark), conda)
        if not os.path.isfile(conda_yaml_path):
            logger.error(
                "ERROR: the conda env file at " + conda_yaml_path + " does not exist"
            )
            sys.exit()
        conda_backend.pin_conda_envs(conda_yaml_path)
    logger.info("DONE: pinned all conda envs.")


## general stuff ######################################################################################


@click.command(no_args_is_help=True)
@click.pass_context
@click.option(
    "-w",
    "--what",
    help="""Binary/functionality to check: \n
               --what singularity : singularity \n
               --what module      : module tool, typically lmod \n
               --what easybuild   : easybuild \n
               --what conda       : conda \n""",
)
def check(ctx, what):
    """Check whether the component {what} is available."""
    logger.info(
        "Checking software stack handlers / backends (singularity, easybuild, etc)."
    )
    from omnibenchmark.software import common

    if what == "easybuild":
        ret = common.check_easybuild_status()
    elif what == "module":
        # eb.export_lmod_env_vars()
        ret = common.check_lmod_status()
    elif what == "singularity":
        ret = common.check_singularity_status()
    elif what == "conda":
        ret = common.check_conda_status()
    # elif what == "docker":
    #     ret = common.check_docker_status()
    else:
        raise click.BadParameter(
            "Bad `--what` value. Please check help (`ob software check --help`)."
        )
    if ret.returncode == 0:
        logger.info("OK: " + ret.stdout)
    else:
        logger.error("Failed: " + ret.stdout + ret.stderr)


software.add_command(check)



#! /usr/bin/env python
# WARNING: Custom dependencies might not be available here, since this is run inside a specified environment.

import logging
import os
import sys
import traceback

from pathlib import Path

from snakemake.script import Snakemake

from omnibenchmark.benchmark import constants
from omnibenchmark.io.code import clone_module
from omnibenchmark.workflow.snakemake.scripts.execution import execution
from omnibenchmark.benchmark.symlinks import SymlinkManager
from omnibenchmark.workflow.snakemake.scripts.utils import (
    get_module_name_from_rule_name,
)

logger = logging.getLogger("SNAKEMAKE_RUNNER")

try:
    snakemake: Snakemake = snakemake
except NameError:
    raise RuntimeError("This script must be run from within a Snakemake workflow")

params = dict(snakemake.params)

repository_url = params.get("repository_url")
commit_hash = params.get("commit_hash")
parameters = params.get("parameters")
inputs_map = params.get("inputs_map")
dataset = params.get("dataset", getattr(snakemake.wildcards, "dataset", "unknown"))

# For now we're handling timeout in seconds.
# When implementing cluster resource handling, we needt to convert this to minutes (e.g. slurm takes it in min)
timeout = params.get(constants.LOCAL_TIMEOUT_VAR, constants.DEFAULT_TIMEOUT_SECONDS)

keep_module_logs = params.get("keep_module_logs", False)
keep_going = snakemake.config.get("keep_going", False)

# For now we're handling timeout in seconds as runtime.
# When implementing cluster resource handling, we will need to convert this to minutes (e.g. slurm takes it in min)
timeout = params.get(constants.LOCAL_TIMEOUT_VAR, constants.DEFAULT_TIMEOUT_SECONDS)

output_dir = Path(str(os.path.commonpath(snakemake.output)))
if len(snakemake.output) == 1:
    output_dir = Path(os.path.dirname(output_dir))

manager = SymlinkManager(output_dir.parent)
manager.store(parameters)

# Clone git repository
repositories_dir = Path(".snakemake") / "repos"
module_dir = clone_module(repositories_dir, repository_url, commit_hash)

# Execute module code
module_name = get_module_name_from_rule_name(snakemake.rule)

try:
    exit_code = execution(
        module_dir,
        module_name=module_name,
        output_dir=output_dir,
        dataset=dataset,
        inputs_map=inputs_map,
        parameters=parameters,
        keep_module_logs=keep_module_logs,
        timeout=timeout,
    )
    if exit_code != 0:
        logger.warning(f"Module execution failed with exit code {exit_code}")
        if not keep_going:
            sys.exit(1)

except Exception as e:
    traceback.print_exc()
    logger.error(f"CRITICAL: {e}")
    if not keep_going:
        # explicit raise will leave a clearer traceback
        raise

sys.exit(0)



import os.path
from collections import Counter
from pathlib import Path
from typing import Optional
from urllib.parse import urlparse

from omni_schema.datamodel.omni_schema import SoftwareBackendEnum, SoftwareEnvironment

from omnibenchmark.benchmark.converter import LinkMLConverter
from omnibenchmark.benchmark.validation.error import ValidationError
from omnibenchmark.utils import try_avail_envmodule


class Validator:
    """Simple validator class for Benchmark."""

    def __init__(self):
        self.errors = []

    def validate(
        self, benchmark_dir: Path, converter: LinkMLConverter
    ) -> LinkMLConverter:
        # Validate ids are unique
        stage_ids = converter.get_stages().keys()
        duplicate_stage_ids = Validator.find_duplicate(stage_ids)
        if duplicate_stage_ids:
            self.errors.append(
                ValidationError(
                    f"Found duplicate stage ids: {', '.join(duplicate_stage_ids)}"
                )
            )

        module_ids = converter.get_modules().keys()
        duplicate_module_ids = Validator.find_duplicate(module_ids)
        if duplicate_module_ids:
            self.errors.append(
                ValidationError(
                    f"Found duplicate module ids: {', '.join(duplicate_module_ids)}"
                )
            )

        output_ids = converter.get_outputs().keys()
        duplicate_output_ids = Validator.find_duplicate(output_ids)
        if duplicate_output_ids:
            self.errors.append(
                ValidationError(
                    f"Found duplicate output ids: {', '.join(duplicate_output_ids)}"
                )
            )

        for stage_id in converter.get_stages():
            stage_inputs_set = converter.get_stage_implicit_inputs(stage_id)
            for stage_inputs in stage_inputs_set:
                for stage_input_id in stage_inputs:
                    if stage_input_id not in output_ids:
                        self.errors.append(
                            ValidationError(
                                f"Input with id '{stage_input_id}' in stage '{stage_id}' is not valid"
                            )
                        )

        # Validate software environments
        software_environments = converter.get_software_environments()
        for module_id, module in converter.get_modules().items():
            environment_id = module.software_environment
            if software_environments.get(environment_id) is None:
                self.errors.append(
                    ValidationError(
                        f"Software environment with id '{environment_id}' is not defined."
                    )
                )

        software_backend = converter.get_software_backend()
        for environment in software_environments.values():
            if software_backend == SoftwareBackendEnum.host:
                continue

            elif software_backend == SoftwareBackendEnum.envmodules:
                load_result = try_avail_envmodule(environment.envmodule)
                if not load_result:
                    self.errors.append(
                        ValidationError(
                            f"Software environment with id '{environment.id}' and name '{environment.envmodule}' could not be loaded as a valid `envmodule`."
                        )
                    )
            elif (
                software_backend == SoftwareBackendEnum.conda
                or software_backend == SoftwareBackendEnum.docker
                or software_backend == SoftwareBackendEnum.apptainer
            ):
                environment_path = Validator.get_environment_path(
                    software_backend, environment, benchmark_dir
                )

                if not environment_path:
                    self.errors.append(
                        ValidationError(
                            f"Software environment with id '{environment.id}' does not have a valid backend definition for: '{software_backend.text}'."
                        )
                    )

                # else for other backends, attempt to find path on disk
                elif not Validator.is_url(environment_path) and not os.path.exists(
                    environment_path
                ):
                    self.errors.append(
                        ValidationError(
                            f"Software environment path for '{software_backend.text}' does not exist: '{environment_path}'."
                        )
                    )

        # Validate metric collectors
        metric_collectors = converter.get_metric_collectors()
        for collector in metric_collectors:
            collector_id = collector.id

            for collector_input in collector.inputs:
                if collector_input.id not in output_ids:
                    self.errors.append(
                        ValidationError(
                            f"Input with id '{collector_input.id}' for metric collector '{collector_id}' is not valid."
                        )
                    )

            environment_id = collector.software_environment
            if software_environments.get(environment_id) is None:
                self.errors.append(
                    ValidationError(
                        f"Software environment with id '{environment_id}'for metric collector '{collector_id}' is not defined."
                    )
                )

                # Raise ValidationError if there are errors
        if self.errors:
            raise ValidationError(self.errors)
        else:
            return converter

    @staticmethod
    def find_duplicate(ids):
        id_counts = Counter(ids)
        duplicate_ids = [id for id, count in id_counts.items() if count > 1]

        return duplicate_ids

    @staticmethod
    def get_environment_path(
        software_backend: SoftwareBackendEnum,
        software: SoftwareEnvironment,
        benchmark_dir: Path,
    ) -> Optional[str]:
        environment = None
        if (
            software_backend == SoftwareBackendEnum.apptainer
            or software_backend == SoftwareBackendEnum.docker
        ):
            environment = software.apptainer

        elif software_backend == SoftwareBackendEnum.conda:
            environment = software.conda

        elif software_backend == SoftwareBackendEnum.envmodules:
            environment = software.envmodule

        if not environment:
            return None

        if Validator.is_url(environment) or Validator.is_absolute_path(environment):
            environment_path = environment
        elif software_backend == SoftwareBackendEnum.envmodules:
            environment_path = environment
        else:
            environment_path = os.path.join(benchmark_dir, environment)

        return environment_path

    @staticmethod
    def is_url(string: str) -> bool:
        # Check if the string is a valid URL using urlparse
        try:
            result = urlparse(string)
            return all(
                [result.scheme, result.netloc]
            )  # Valid if both scheme and netloc are present
        except ValueError:
            return False

    @staticmethod
    def is_absolute_path(string: str) -> bool:
        # Check if the string is an absolute path
        return Path(string).is_absolute()



"""MinIO class for remote storage."""

import datetime
import io
import json
import logging
import os
import re

from packaging.version import Version
from pathlib import Path
from typing import Dict, Union
from urllib.parse import urlparse

import boto3
import minio
import minio.commonconfig
import minio.retention

from omnibenchmark.benchmark import Benchmark
from omnibenchmark.io.exception import (
    MinIOStorageBucketManipulationException,
    MinIOStorageConnectionException,
    RemoteStorageInvalidInputException,
)
from omnibenchmark.io.archive import prepare_archive_software
from omnibenchmark.io.RemoteStorage import RemoteStorage, StorageOptions
from omnibenchmark.io.S3config import bucket_readonly_policy
from omnibenchmark.io.S3versioning import get_s3_object_versions_and_tags
from omnibenchmark.io.versioning import (
    filter_objects_to_tag,
    get_objects_to_tag,
    get_remoteversion_from_bmversion,
    prepare_csv_remoteversion_from_bmversion,
)

# TODO: set if we have the DEBUG flag set
logging.getLogger("requests").setLevel(logging.DEBUG)
logging.getLogger("minio").setLevel(logging.DEBUG)
logger = logging.getLogger(__name__)


def set_bucket_public_readonly(client: minio.Minio, bucket_name: str):
    policy = bucket_readonly_policy(bucket_name)
    client.set_bucket_policy(bucket_name, json.dumps(policy))


def set_bucket_lifecycle_config(
    client: minio.Minio, bucket_name: str, noncurrent_days: int = 1
):
    lifecycle_config = minio.lifecycleconfig.LifecycleConfig(
        [
            minio.lifecycleconfig.Rule(
                minio.commonconfig.ENABLED,
                rule_filter=minio.lifecycleconfig.Filter(prefix="*"),
                rule_id="rule1",
                noncurrent_version_expiration=minio.lifecycleconfig.NoncurrentVersionExpiration(
                    noncurrent_days=noncurrent_days
                ),
            ),
        ],
    )
    client.set_bucket_lifecycle(bucket_name, lifecycle_config)


class MinIOStorage(RemoteStorage):
    def __init__(
        self,
        auth_options: Dict,
        benchmark: str,
        storage_options: StorageOptions,
    ):
        super().__init__(auth_options, benchmark, storage_options)
        assert "endpoint" in self.auth_options.keys()
        if "access_key" in self.auth_options.keys():
            self.client = self.connect()
            self.roclient = self.connect(readonly=True)
            self._test_connect()

            if not self.client.bucket_exists(benchmark):
                logger.warning(
                    f"Benchmark {benchmark} does not exist, creating new benchmark."
                )
                self._create_benchmark(benchmark)

            self._get_versions()
        else:
            self.roclient = self.connect(readonly=True)
            self._get_versions()

    def connect(self, readonly=False) -> minio.Minio:
        """
        Connects to the MinIO storage.

        Returns:
        - A MinIO client object.
        """
        if readonly:
            assert "endpoint" in self.auth_options.keys()
            tmp_auth_options = self.auth_options.copy()
            if "access_key" in tmp_auth_options.keys():
                del tmp_auth_options["access_key"]
            if "secret_key" in tmp_auth_options.keys():
                del tmp_auth_options["secret_key"]
        else:
            assert (
                "endpoint" in self.auth_options.keys()
                and "access_key" in self.auth_options.keys()
                and "secret_key" in self.auth_options.keys()
            )
            tmp_auth_options = self.auth_options.copy()
        try:
            return minio.Minio(**tmp_auth_options)
        except Exception:
            url = urlparse(tmp_auth_options["endpoint"])
            tmp_auth_options["endpoint"] = url.netloc
            return minio.Minio(**tmp_auth_options)

    def _test_connect(self) -> None:
        try:
            _ = self.client.list_objects(self.benchmark)
        except MinIOStorageConnectionException as e:
            raise e

    def _create_benchmark(self, benchmark: str) -> None:
        if self.client.bucket_exists(benchmark):
            raise MinIOStorageBucketManipulationException(
                f"Benchmark {benchmark} already exists."
            )
        # create new version
        self.client.make_bucket(bucket_name=benchmark, object_lock=True)
        if self.client._base_url.is_aws_host:
            s3 = boto3.client(
                "s3",
                aws_access_key_id=self.auth_options["access_key"],
                aws_secret_access_key=self.auth_options["secret_key"],
            )
            s3.delete_public_access_block(Bucket=benchmark)
        set_bucket_public_readonly(self.client, benchmark)
        set_bucket_lifecycle_config(self.client, benchmark, noncurrent_days=1)
        if not self.client.bucket_exists(benchmark):
            raise MinIOStorageBucketManipulationException(
                f"Bucket creation for benchmark {benchmark} failed"
            )

    def _get_versions(self) -> None:
        versionobjects = list(
            self.roclient.list_objects(
                self.benchmark, prefix="versions", recursive=True
            )
        )
        allversions = [
            os.path.basename(v.object_name).replace(".csv", "") for v in versionobjects
        ]
        versions = list()
        for version in allversions:
            if re.match("(\\d+).(\\d+)", version):
                versions.append(Version(version))
        self.versions = versions

    def create_new_version(self, benchmark: Union[None, Benchmark] = None) -> None:
        if self.version is None:
            raise RemoteStorageInvalidInputException(
                "No version provided, set version first with method 'set_version'"
            )

        if "access_key" not in self.auth_options.keys():
            raise RemoteStorageInvalidInputException(
                "Read-only mode, cannot create new version, set access_key and secret_key in auth_options"
            )

        # update
        self._get_versions()
        if self.version in self.versions:
            raise RemoteStorageInvalidInputException(
                "Version already exists, set new version with method 'set_version'"
            )

        # upload benchmark definition file
        if benchmark is not None:
            # read file
            bmfile = Path(benchmark.get_definition_file())
            with open(bmfile, "r") as fh:
                bmstr = fh.read()

            # upload file
            _ = self.client.put_object(
                self.benchmark,
                "config/benchmark.yaml",
                io.BytesIO(bmstr.encode()),
                len(bmstr),
            )

            # read software files
            software_files = prepare_archive_software(benchmark)

            # upload software files
            for software_file in software_files:
                with open(software_file, "r") as fh:
                    softstr = fh.read()

                _ = self.client.put_object(
                    self.benchmark,
                    f"software/{software_file.name}",
                    io.BytesIO(softstr.encode()),
                    len(softstr),
                )

        # get all objects
        objdic = get_s3_object_versions_and_tags(self.client, self.benchmark)

        # get objects to tag
        object_names_to_tag, versionid_of_objects_to_tag = get_objects_to_tag(
            objdic, storage_options=self.storage_options
        )

        # filter objects based on workflow
        object_names_to_tag, versionid_of_objects_to_tag = filter_objects_to_tag(object_names_to_tag,
                                                                                 versionid_of_objects_to_tag,
                                                                                 self.storage_options,
                                                                                 benchmark)

        # Tag all objects with current version
        tags = minio.datatypes.Tags.new_object_tags()
        tags[str(self.version)] = "1"
        for n, v in zip(object_names_to_tag, versionid_of_objects_to_tag):
            self.client.set_object_tags(self.benchmark, n, tags, version_id=v)

        # set retention policy of objects
        retention_config = minio.retention.Retention(
            minio.commonconfig.GOVERNANCE,
            datetime.datetime.now(datetime.UTC) + datetime.timedelta(weeks=1000),
        )
        for n, v in zip(object_names_to_tag, versionid_of_objects_to_tag):
            self.client.set_object_retention(
                self.benchmark, n, config=retention_config, version_id=v
            )

        # refresh object list
        objdic = get_s3_object_versions_and_tags(self.client, self.benchmark)

        # write versions/{{version}}.csv
        vv_ls = get_remoteversion_from_bmversion(objdic, str(self.version))
        vv_str = prepare_csv_remoteversion_from_bmversion(vv_ls)
        version_filename = f"versions/{self.version}.csv"
        _ = self.client.put_object(
            self.benchmark, version_filename, io.BytesIO(vv_str.encode()), len(vv_str)
        )

        # tag and set retention policy of version file
        self.client.set_object_tags(self.benchmark, version_filename, tags)
        self.client.set_object_retention(
            self.benchmark, version_filename, config=retention_config
        )

        # check if version exists
        self._get_versions()
        if self.version not in self.versions:
            raise MinIOStorageBucketManipulationException("Version creation failed")

    # TODO(ben): review if we want this public, it's used in tests
    def _get_objects(self) -> None:
        if self.version is None:
            raise RemoteStorageInvalidInputException(
                "No version provided, set version first with method 'set_version'"
            )

        if self.version in self.versions:
            # read overview file
            response = self.roclient.get_object(
                self.benchmark, f"versions/{self.version}.csv"
            )
            objls = response.data.decode("utf-8")
            objls = objls.split("\n")
            objls = [obj for obj in objls if obj]
            objdict = {}
            header = objls[0].split(",")
            assert header[0] == "name"
            for obj in objls[1:]:
                tmpsplit = obj.split(",")
                objdict[tmpsplit[0]] = {}
                for i, head in enumerate(header[1:]):
                    objdict[tmpsplit[0]][head] = tmpsplit[i + 1]

            # add overview file to files
            response_headers = response.headers
            objdict[f"versions/{self.version}.csv"] = {
                "version_id": response_headers.get("x-amz-version-id"),
                # some parsing of date to get to consistent format
                "last_modified": datetime.datetime.strptime(
                    response_headers.get("last-modified"), "%a, %d %b %Y %H:%M:%S GMT"
                ).strftime("%Y-%m-%d %H:%M:%S.%f+00:00"),
                "size": response_headers.get("content-length"),
                "etag": response_headers.get("etag").replace('"', ""),
            }
        else:
            # get all objects
            objdic = get_s3_object_versions_and_tags(
                self.roclient, self.benchmark, readonly=True
            )

            # get objects to tag
            object_names_to_tag, versionid_of_objects_to_tag = get_objects_to_tag(
                objdic, storage_options=self.storage_options
            )
            objdict = {}
            for obj, vt in zip(object_names_to_tag, versionid_of_objects_to_tag):
                objdict[obj] = {}
                objdict[obj]["version_id"] = vt
                objdict[obj]["etag"] = objdic[obj][vt]["etag"]
                objdict[obj]["size"] = objdic[obj][vt]["size"]
                objdict[obj]["last_modified"] = objdic[obj][vt]["last_modified"]

        self.files = objdict

    def download_object(self, object_name: str, local_path: str) -> None:
        if self.version is None:
            raise RemoteStorageInvalidInputException(
                "No version provided. Set version first with method 'set_version'"
            )
        if self.files is None:
            self._get_objects()
        if self.files is None:
            raise RemoteStorageInvalidInputException("No objects found")
        if object_name not in self.files.keys():
            raise RemoteStorageInvalidInputException(f"Object {object_name} not found")
        _ = self.roclient.fget_object(
            self.benchmark,
            object_name,
            local_path,
            version_id=self.files[object_name]["version_id"],
        )

    def archive_version(
        self,
        benchmark: Benchmark,
        outdir: Path = Path(),
        config: bool = True,
        code: bool = False,
        software: bool = False,
        results: bool = False,
    ):
        from omnibenchmark.io.archive import archive_version

        # TODO: upload the zip archive
        _ = archive_version(benchmark, outdir, config, code, software, results)

    def delete_version(self, version):
        raise NotImplementedError


RemoteStorage.register(MinIOStorage)



import json
import os
import sys


from omnibenchmark.cli.utils.logging import logger


def benchmarker_access_token_policy(benchmark):
    """S3 policy for access token for specific benchmark, allows archiving (Governance Retention)"""
    return {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Action": ["s3:*"],
                "Resource": [
                    f"arn:aws:s3:::{benchmark}/*",
                    f"arn:aws:s3:::{benchmark}",
                ],
            },
            {
                "Effect": "Deny",
                "Action": [
                    "s3:BypassGovernanceRetention",
                    "s3:DeleteObjectTagging",
                    "s3:DeleteObjectVersion",
                    "s3:DeleteObjectVersionTagging",
                    "s3:DeleteBucket",
                    "s3:DeleteBucketPolicy",
                ],
                "Resource": [
                    f"arn:aws:s3:::{benchmark}/*",
                    f"arn:aws:s3:::{benchmark}",
                ],
            },
        ],
    }


def bucket_readonly_policy(bucket_name):
    return {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Principal": {"AWS": "*"},
                "Action": ["s3:GetBucketLocation", "s3:ListBucket"],
                "Resource": f"arn:aws:s3:::{bucket_name}",
            },
            {
                "Effect": "Allow",
                "Principal": {"AWS": "*"},
                "Action": "s3:GetObject",
                "Resource": f"arn:aws:s3:::{bucket_name}/*",
            },
        ],
    }


def S3_access_config_from_env() -> dict:
    """Get S3 access config from environment variables or file"""
    if (
        "OB_STORAGE_S3_ACCESS_KEY" in os.environ
        and "OB_STORAGE_S3_SECRET_KEY" in os.environ
    ):
        return {
            "access_key": os.environ["OB_STORAGE_S3_ACCESS_KEY"],
            "secret_key": os.environ["OB_STORAGE_S3_SECRET_KEY"],
        }
    elif "OB_STORAGE_S3_CONFIG" in os.environ:
        with open(os.environ["OB_STORAGE_S3_CONFIG"], "r") as file:
            auth_options = json.load(file)
        if "access_key" in auth_options and "secret_key" in auth_options:
            return auth_options
        else:
            logger.error(
                f"Invalid S3 config, missing access_key or secret_key in config file ({os.environ['OB_STORAGE_S3_CONFIG']})",
            )
            sys.exit(1)
    else:
        logger.error(
            "Invalid S3 config. Missing access_key and secret_key in environment variables (OB_STORAGE_S3_ACCESS_KEY, OB_STORAGE_S3_SECRET_KEY) or OB_STORAGE_S3_CONFIG",
        )
        sys.exit(1)


if __name__ == "__main__":
    print(json.dumps(benchmarker_access_token_policy("obob"), indent=2))
    print(json.dumps(bucket_readonly_policy("omnibenchmarktestbucket"), indent=2))



from collections import OrderedDict
from collections.abc import Mapping
import json
import hashlib


class Params:
    """
    Params is a data structure for workflow steps that provides ordered storage,
    serialization, and stable hashing capabilities.
    """

    def __init__(self, params=None):
        """Initialize with optional parameters dict."""
        self._params = OrderedDict()
        if params:
            # Convert input to ordered dict, sorting by keys
            if isinstance(params, (dict, OrderedDict)):
                for k in sorted(params.keys()):
                    self._params[k] = params[k]
            elif isinstance(params, Params):
                self._params = params._params.copy()

    def __eq__(self, other):
        """
        Compare for equality with another object.
        Handles comparison with other Params objects and dict-like objects.
        """
        if isinstance(other, Params):
            return self.serialize() == other.serialize()
        elif isinstance(other, Mapping):
            # Convert dict-like object to Params and compare
            return self == Params(other)
        return NotImplemented

    def __getitem__(self, key):
        """Allow dict-like access."""
        return self._params[key]

    def __setitem__(self, key, value):
        """Allow dict-like assignment."""
        self._params[key] = value

    def __str__(self):
        """Return canonical string representation."""
        return self.serialize()

    def __repr__(self):
        """Return canonical string representation."""
        return self.__str__()

    def hash(self):
        """Generate stable hash of canonical form."""
        canonical = self.serialize()
        return hashlib.sha256(canonical.encode()).hexdigest()

    def items(self):
        """Return items iterator."""
        return self._params.items()

    def update(self, other):
        """Update params from dict or another Params object."""
        if isinstance(other, Params):
            self._params.update(other._params)
        else:
            self._params.update(other)

    def serialize(self):
        """Convert to canonical JSON string."""
        # Sort keys to ensure canonical form
        return json.dumps(self._params, sort_keys=True)

    @classmethod
    def deserialize(cls, serialized):
        """Create new Params instance from serialized string."""
        params = json.loads(serialized)
        return cls(params)

    def to_cli_args(self, style="gnu"):
        """
        Convert parameters to command line arguments.

        Args:
            style (str): Either 'gnu' (--key value) or
                        'equals' (--key=value)

        Returns:
            list: List of command line argument strings
        """
        if style not in ["gnu", "equals"]:
            raise ValueError("style must be either 'gnu' or 'equals'")

        args = []
        for key, value in self._params.items():
            if isinstance(value, bool) and value:
                args.append(f"--{key}")
            else:
                if style == "gnu":
                    args.extend([f"--{key}", str(value)])
                else:  # equals style
                    args.append(f"--{key}={value}")
        return args

    @classmethod
    def from_cli_args(cls, args):
        """
        Create Params from command line arguments.
        Supports:
        - Long options: --key value and --key=value
        - Short options: -k value and -k=value
        - Implicit flags: --key and -k (True)

        Args:
            args (list): List of command line argument strings

        Returns:
            Params: New Params instance
        """
        params = {}
        i = 0
        while i < len(args):
            arg = args[i]

            # If arg contains a space inside itself ("-a 0")
            if " " in arg:
                first_part, second_part = arg.split(maxsplit=1)
                if first_part.startswith("--") or first_part.startswith("-"):
                    prefix_len = 2 if first_part.startswith("--") else 1
                    key = first_part[prefix_len:]
                    value = second_part
                    params[key] = value
                i += 1
                continue

            if arg.startswith("--") or arg.startswith("-"):
                # Remove -- or - prefix
                prefix_len = 2 if arg.startswith("--") else 1
                key = arg[prefix_len:]

                # Handle --key=value or -k=value format
                if "=" in key:
                    key, value = key.split("=", 1)
                    params[key] = value
                    i += 1
                    continue

                # Handle --key value or -k value format
                if i + 1 < len(args) and not args[i + 1].startswith("-"):
                    params[key] = args[i + 1]
                    i += 2
                else:
                    # Implicit flags (no value means True)
                    params[key] = True
                    i += 1
            else:
                # Skip non-option arguments
                i += 1

        return cls(params)



from pathlib import Path
from typing import List, Tuple

import networkx as nx
import pydot

import omni_schema.datamodel.omni_schema

from omnibenchmark.benchmark.converter import LinkMLConverter
from omnibenchmark.benchmark.benchmark_node import BenchmarkNode
from omnibenchmark.benchmark.validation import ValidationError


def expend_stage_nodes(
    converter: LinkMLConverter,
    stage: omni_schema.datamodel.omni_schema.Stage,
    out_dir: Path,
    stage_ordering: List[str],
) -> List[BenchmarkNode]:
    nodes = []

    input_dirname = str(out_dir) if converter.is_initial(stage) else "{pre}"
    stage_outputs = converter.get_stage_outputs(stage).values()
    outputs = [x.replace("{input}", input_dirname) for x in stage_outputs]

    inputs_for_stage = converter.get_stage_implicit_inputs(stage)
    if not inputs_for_stage or len(inputs_for_stage) == 0:
        inputs_for_stage = [None]

    modules_in_stage = converter.get_modules_by_stage(stage)
    for module_id in modules_in_stage:
        module = modules_in_stage[module_id]
        parameters = converter.get_module_parameters(module)
        if not parameters or len(parameters) == 0:
            parameters = [None]

        for param in parameters:
            for inputs in inputs_for_stage:
                latest_stage = None
                if inputs:
                    stages_by_output = list(
                        set(
                            [
                                converter.get_stage_by_output(input_id).id
                                for input_id in inputs
                            ]
                        )
                    )
                    latest_stage = sorted(stages_by_output, key=stage_ordering.index)[
                        -1
                    ]
                    explicit_inputs = converter.get_explicit_inputs(inputs)
                    inputs = {
                        k: v.replace("{input}", "{pre}")
                        for k, v in explicit_inputs.items()
                    }

                node = BenchmarkNode(
                    converter,
                    stage,
                    module,
                    param,
                    inputs,
                    outputs,
                    after=latest_stage,
                )
                nodes.append(node)

    return nodes


def build_benchmark_dag(converter: LinkMLConverter, out_dir: Path) -> nx.DiGraph:
    g = nx.DiGraph()

    G_stages = build_stage_dag(converter)
    stage_ordering = compute_stage_order(G_stages)

    stage_nodes_map = {}
    for stage_id, stage in converter.get_stages().items():
        nodes = expend_stage_nodes(converter, stage, out_dir, stage_ordering)
        nodes_with_stage = [(node, {"stage": stage_id}) for node in nodes]
        g.add_nodes_from(nodes_with_stage)
        stage_nodes_map[stage_id] = nodes

    for current_node in g.nodes:
        after_stage = current_node.after
        if after_stage:
            departure_stage_nodes = stage_nodes_map[after_stage]
            for departure_stage_node in departure_stage_nodes:
                g.add_edge(departure_stage_node, current_node)

    return g


def build_stage_dag(converter: LinkMLConverter) -> nx.DiGraph:
    g = nx.DiGraph()

    for stage_id, stage in converter.get_stages().items():
        g.add_node(stage_id)
        input_ids = [input_id for input in stage.inputs for input_id in input.entries]
        dep_stages = [converter.get_output_stage(input_id) for input_id in input_ids]
        for dep in dep_stages:
            g.add_edge(dep.id, stage.id)

    return g


def find_initial_and_terminal_nodes(
    graph: nx.DiGraph,
) -> Tuple[List[BenchmarkNode], List[BenchmarkNode]]:
    initial_nodes = [node for node, in_degree in graph.in_degree() if in_degree == 0]
    terminal_nodes = [
        node for node, out_degree in graph.out_degree() if out_degree == 0
    ]
    return initial_nodes, terminal_nodes


def list_all_paths(graph: nx.DiGraph, source: BenchmarkNode, target: BenchmarkNode):
    all_paths = list(nx.all_simple_paths(graph, source=source, target=target))
    return all_paths


def contains_all(path, modules):
    path_modules = [node.module_id for node in path]
    return all(module in path_modules for module in modules)


def exclude_paths(paths, path_exclusions):
    updated_paths = []
    for path in paths:
        should_exclude = False
        for module, excluded_modules in path_exclusions.items():
            for excluded_module in excluded_modules:
                if contains_all(path, [module, excluded_module]):
                    should_exclude = True

        if not should_exclude:
            updated_paths.append(path)

    return updated_paths


def compute_stage_order(stage_dag: nx.DiGraph) -> List:
    try:
        topological_order = list(nx.topological_sort(stage_dag))

    except nx.NetworkXUnfeasible:
        raise ValidationError(
            "The stage graph has a cyclic dependencies. This benchmark can not be resolved"
        )

    return topological_order


def export_to_dot(
    G: nx.DiGraph,
    title: str = None,
):
    import matplotlib.pyplot as plt

    # Dynamically scale the node size based on node count
    nodes_count = len(G.nodes)
    div_nodes_count = max(1, nodes_count // 10)
    graph_size = max(15, 15 * div_nodes_count)

    # Color nodes by stage (assuming 'stage' is a node attribute)
    stages = nx.get_node_attributes(G, "stage", default="none")
    unique_stages = list(set(stages.values()))  # Get unique stages

    # Define a colormap with different shades for the stages
    stage_colors = plt.get_cmap("inferno", max(len(unique_stages), 5))

    # Convert the graph to a PyDot graph object
    pydot_graph = pydot.Dot(
        graph_type="digraph", strict=True, label=title, labelloc="top", fontsize=20
    )
    pydot_graph.set_graph_defaults(
        size=f"{graph_size},{graph_size}!", ratio="fill", margin=div_nodes_count
    )

    # Define the style for nodes
    node_defaults = {
        "shape": "rect",
        "style": "filled,rounded",
        "fontsize": "12",
        "fontcolor": "white",
        "width": "1.5",
        "height": "0.6",
        "penwidth": "1.0",
    }
    pydot_graph.set_node_defaults(**node_defaults)

    # Define the style for edges
    edge_defaults = {"color": "#CCCCCC", "penwidth": "0.5", "arrowsize": "0.7"}
    pydot_graph.set_edge_defaults(**edge_defaults)

    for node in G.nodes:
        node_name = str(node)
        rgba_color = stage_colors(unique_stages.index(stages[node]))
        hex_color = _rgba_to_hex(rgba_color)
        pydot_node = pydot.Node(node_name, label=node_name, fillcolor=hex_color)
        pydot_graph.add_node(pydot_node)

    for source, target in G.edges:
        pydot_edge = pydot.Edge(str(source), str(target))
        pydot_graph.add_edge(pydot_edge)

    return pydot_graph


def _rgba_to_hex(rgba):
    r = int(rgba[0] * 255)
    g = int(rgba[1] * 255)
    b = int(rgba[2] * 255)
    return f"#{r:02X}{g:02X}{b:02X}"



import logging
import sys


logger = logging.getLogger("omnibenchmark")


def configure_logging(debug: bool):
    """
    Configures the logging system based on the debug flag.
    """
    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter("%(message)s")
    handler.setFormatter(formatter)

    log_level = logging.DEBUG if debug else logging.INFO
    logger.setLevel(log_level)

    if not logger.hasHandlers():
        logger.addHandler(handler)



import hashlib


def get_module_name_from_rule_name(rule_name: str) -> str:
    return rule_name.split("_")[1]


# Create a unique folder name based on the repository URL and commit hash
def generate_unique_repo_folder_name(repo_url: str, commit_hash: str) -> str:
    unique_string = f"{repo_url}@{commit_hash}"
    folder_name = hashlib.sha256(unique_string.encode()).hexdigest()

    return folder_name



"""omnibenchmark CLI"""

import click

from omnibenchmark import __version__
from omnibenchmark.cli.benchmark import info
from omnibenchmark.cli.io import storage
from omnibenchmark.cli.run import run
from omnibenchmark.cli.soft import software

from .debug import add_debug_option


@click.group()
@click.version_option(__version__, prog_name="OmniBenchmark CLI")
@click.pass_context
def cli(ctx):
    """
    OmniBenchmark Command Line Interface (CLI).
    """
    ctx.ensure_object(dict)


# Add subcommands to the CLI
cli.add_command(add_debug_option(storage))
cli.add_command(add_debug_option(software))
cli.add_command(add_debug_option(run))
cli.add_command(add_debug_option(info))

add_debug_option(cli)

if __name__ == "__main__":
    cli(obj={})



import re
from itertools import takewhile
from pathlib import Path
from typing import List, Set, Tuple, Union, NamedTuple, Dict

from omni_schema.datamodel.omni_schema import MetricCollector

from omnibenchmark.benchmark import BenchmarkNode, Benchmark
from omnibenchmark.utils import format_mc_output


class Wildcards(NamedTuple):
    pre: str
    post: str
    dataset: str


def format_performance_file(node: BenchmarkNode, out_dir: str) -> str:
    """Provides a benchmark performance path for a node"""

    is_initial = node.is_initial()
    if is_initial:
        return out_dir + "/{stage}/{module}/{params}/{dataset}_performance.txt"
    else:
        return "{pre}/{post}/{dataset}_performance.txt"


def format_output_templates_to_be_expanded(node: BenchmarkNode) -> List[str]:
    """Formats node outputs that will be expanded according to Snakemake's engine"""
    outputs = node.get_outputs()
    is_initial = node.is_initial()

    if not is_initial:
        outputs = [re.sub(r"\/.*\/", "/{post}/", o, count=1) for o in outputs]

    # print(f'Output: {node.stage_id}: {outputs}')
    return outputs


def format_metric_collector_input(
    benchmark: Benchmark, collector: MetricCollector, return_as_dict=False
) -> Dict[str, str] | List[str]:
    """Formats collector inputs that will be expanded according to Snakemake's engine"""
    implicit_inputs = [i.id for i in collector.inputs]
    explicit_inputs = benchmark.get_explicit_input(implicit_inputs)

    if not return_as_dict:
        inputs = list(explicit_inputs.values())
    else:
        inputs = explicit_inputs

    # print(f'Collector Input: {collector.name}: {inputs}')
    return inputs


def format_metric_collector_output(
    out_dir: Path, collector: MetricCollector
) -> List[str]:
    """Formats collector outputs that will be expanded according to Snakemake's engine"""
    outputs = []
    for o in collector.outputs:
        outputs.append(format_mc_output(o, out_dir, collector.id))

    # print(f'Collector Output: {collector.name}: {outputs}')
    return outputs


def format_input_templates_to_be_expanded(
    benchmark: Benchmark, wildcards: Wildcards, return_as_dict=False
) -> Dict[str, str] | List[str]:
    """Formats benchmark inputs that will be expanded according to Snakemake's engine"""

    pre = wildcards.pre
    post = wildcards.post
    dataset = wildcards.dataset

    nodes = benchmark.get_nodes()
    stage_ids = benchmark.get_stage_ids()

    pre_stages = _extract_stages_from_path(pre, stage_ids)
    matched = _match_node_format(pre_stages[-1]) if len(pre_stages) > 0 else []

    if len(matched) != 3:
        return {} if return_as_dict else []

    after_stage_id, _, _ = matched
    stage_id, module_id, param_id = _match_node_format(post)

    node_hash = hash(BenchmarkNode.to_id(stage_id, module_id, param_id, after_stage_id))
    matching_node = next((node for node in nodes if hash(node) == node_hash), None)
    if not matching_node:
        return {} if return_as_dict else []

    node_inputs = matching_node.get_inputs_dict()

    inputs = _match_inputs(node_inputs, pre_stages, pre, dataset)

    if return_as_dict:
        return inputs
    else:
        return inputs.values()


def _extract_stages_from_path(
    path: str, known_stage_ids: Set[str]
) -> List[Union[str, tuple]]:
    def is_known_stage(part: str) -> bool:
        return part in known_stage_ids

    def find_sub_parts(start_index: int) -> Union[str, tuple]:
        sub_parts = list(
            takewhile(lambda x: x not in known_stage_ids, parts[start_index + 1 :])
        )
        return tuple(parts[start_index : start_index + 1 + len(sub_parts)])

    if not path:
        return []

    parts = [part for part in Path(path).parts]
    stages = [
        find_sub_parts(part_idx)
        for part_idx, part in enumerate(parts)
        if is_known_stage(part)
    ]

    return stages


def _match_node_format(to_match: Union[str, tuple]) -> Tuple[str, str, str]:
    if isinstance(to_match, str):
        # TODO(daninci): Fix type checking
        to_match = [part for part in Path(to_match).parts]

    stage_id = to_match[0]
    module_id = to_match[1]
    param_id = to_match[2]

    return stage_id, module_id, param_id


def _match_input_module(input: str, stages: List[Tuple[str]], dataset: str) -> str:
    expected_input_module = input.split("{pre}/")[1].split("/{module}")[0]
    matching_stage = next(
        (tup for tup in stages if tup[0] == expected_input_module), None
    )
    if not matching_stage:
        raise RuntimeError(f"Could not find matching stage for {input} in {stages}")

    if len(matching_stage) < 3:
        raise RuntimeError(f"{stages} has wrong length")

    assert len(matching_stage) >= 3
    matched_module = matching_stage[1]

    input = input.replace("{module}", matched_module)
    input = input.replace("{dataset}", dataset)
    if "{params}" in input:
        matched_params = next(
            (x for x in matching_stage[2:] if "param" or "default" in x), None
        )
        input = input.replace("{params}", matched_params)

    return input


def _match_input_prefix(input: str, pre: str) -> str:
    stage = f"/{input.split('/')[1]}"
    matched_prefix = pre.split(stage)[0]
    formatted_input = input.format(pre=matched_prefix)

    # print(f'Input: {input} Prefix: {pre} Formatted: {formatted_input}')
    return formatted_input


def _match_inputs(
    inputs: dict[str, str], stages: List[Tuple[str]], pre: str, dataset: str
) -> dict[str, str]:
    all_matched = True

    formatted_inputs = {}
    for key, input in inputs.items():
        formatted_input = _match_input_module(input, stages, dataset)
        if not formatted_input:
            all_matched = False
            break
        else:
            formatted_input = _match_input_prefix(formatted_input, pre)
            formatted_inputs[key] = formatted_input

    return formatted_inputs if all_matched else {}



"""cli commands related to input/output files"""

import json
import sys
import zipfile

from pathlib import Path

import click
import yaml

from omnibenchmark.benchmark import Benchmark
from omnibenchmark.cli.utils.logging import logger
from omnibenchmark.cli.utils.validation import validate_benchmark
from omnibenchmark.io.archive import archive_version
from omnibenchmark.io.files import checksum_files
from omnibenchmark.io.files import list_files
from omnibenchmark.io.files import download_files
from omnibenchmark.io.S3config import benchmarker_access_token_policy
from omnibenchmark.io.tree import tree_string_from_list
from omnibenchmark.io.storage import get_storage, remote_storage_args

from .debug import add_debug_option


@click.group(name="storage")
@click.pass_context
def storage(ctx):
    """Manage remote storage."""
    ctx.ensure_object(dict)


@add_debug_option
@storage.command(name="create-version")
@click.option(
    "-b",
    "--benchmark",
    help="Path to benchmark yaml file or benchmark id.",
    type=click.Path(exists=True),
    required=True,
    envvar="OB_BENCHMARK",
)
def create_benchmark_version(benchmark: str):
    """Create a new benchmark version."""

    with open(benchmark, "r") as fh:
        yaml.safe_load(fh)
        benchmark = Benchmark(Path(benchmark))

    auth_options = remote_storage_args(benchmark)

    # setup storage
    ss = get_storage(
        str(benchmark.converter.model.storage_api),
        auth_options,
        str(benchmark.converter.model.storage_bucket_name),
    )
    ss.set_version(benchmark.get_benchmark_version())
    if ss.version in ss.versions:
        logger.error(
            "Error: version already exists. Cannot overwrite.",
        )
        sys.exit(1)
    else:
        logger.info("Create a new benchmark version")
        ss.create_new_version(benchmark)


@add_debug_option
@storage.command(name="list")
@click.option(
    "-b",
    "--benchmark",
    help="Path to benchmark yaml file or benchmark id.",
    type=click.Path(exists=True),
    required=True,
    envvar="OB_BENCHMARK",
)
@click.option(
    "-t",
    "--type",
    help="File types. Options: all, code, inputs, outputs, logs, performance.",
    type=str,
    default="all",
)
@click.option("-s", "--stage", help="Stage to list files for.", type=str, default=None)
@click.option(
    "-i", "--id", "file_id", help="File id/type to list.", type=str, default=None
)
def list_all_files(
    benchmark: str,
    type: str = "all",
    stage: str = "",
    module: str = "",
    file_id: str = "",
):
    """List all or specific files for a benchmark."""
    if file_id is not None:
        logger.error("--file_id is not implemented")
        sys.exit(1)
    if type != "all":
        logger.error("--type is not implemented")
        sys.exit(1)

    objectnames, etags = list_files(benchmark, type, stage, module, file_id)
    if len(objectnames) > 0:
        for objectname, etag in zip(objectnames, etags):
            logger.info(f"{etag} {objectname}")


@add_debug_option
@storage.command(name="download")
@click.option(
    "-b",
    "--benchmark",
    help="Path to benchmark yaml file or benchmark id.",
    type=click.Path(exists=True),
    required=True,
    envvar="OB_BENCHMARK",
)
@click.option(
    "-t",
    "--type",
    help="File types. Options: all, code, inputs, outputs, logs, performance.",
    type=str,
    default="all",
)
@click.option(
    "-s", "--stage", help="Stage to download files from.", type=str, default=None
)
@click.option(
    "-m", "--module", help="Module to download files from.", type=str, default=None
)
@click.option(
    "-i", "--id", "file_id", help="File id to download.", type=str, default=None
)
@click.option(
    "-o",
    "--overwrite",
    help="Overwrite existing files.",
    is_flag=True,
    default=False,
    show_default=True,
)
def download_all_files(
    benchmark: str,
    type: str = "all",
    stage: str = "",
    module: str = "",
    file_id: str = "",
    overwrite: bool = False,
):
    """Download all or specific files for a benchmark."""
    if file_id is not None:
        logger.error("--file_id is not implemented")
        raise click.Abort()
    if type != "all":
        logger.info("--type is not implemented")
        raise click.Abort()

    download_files(
        benchmark,
        type,
        stage,
        module,
        file_id,
        verbose=True,
        overwrite=overwrite,
    )


@add_debug_option
@storage.command(name="checksum")
@click.option(
    "-b",
    "--benchmark",
    help="Path to benchmark yaml file or benchmark id.",
    type=click.Path(exists=True),
    required=True,
    envvar="OB_BENCHMARK",
)
def checksum_all_files(benchmark: str):
    """Generate md5sums of all benchmark outputs"""

    logger.info("Checking MD5 checksums... ")
    failed_checks_filenames = checksum_files(benchmark=benchmark, verbose=True)
    if len(failed_checks_filenames) > 0:
        logger.error("Failed checksums:")
        for filename in failed_checks_filenames:
            logger.error(filename)
        raise click.Abort()
    logger.info("Done")


@storage.command(name="create-policy")
@click.option(
    "-b",
    "--benchmark",
    help="Path to benchmark yaml file or benchmark id.",
    type=click.Path(exists=True),
    required=True,
    envvar="OB_BENCHMARK",
)
def create_policy(benchmark: str):
    """Create a new policy for a benchmark."""

    with open(benchmark, "r") as fh:
        yaml.safe_load(fh)
        benchmark = Benchmark(Path(benchmark))

    if (
        str(benchmark.converter.model.storage_api).upper() == "MINIO"
        or str(benchmark.converter.model.storage_api).upper() == "S3"
    ):
        policy = benchmarker_access_token_policy(
            benchmark.converter.model.storage_bucket_name
        )
        logger.error(json.dumps(policy, indent=2))
    else:
        logger.error("Error: Invalid storage type. Only MinIO/S3 storage is supported.")
        raise click.Abort()


@storage.command("archive")
@click.option(
    "--benchmark",
    "-b",
    required=True,
    type=click.Path(exists=True),
    help="Path to benchmark yaml file or benchmark id.",
    envvar="OB_BENCHMARK",
)
@click.option(
    "-c",
    "--code",
    help="Archive benchmarking code (repos).",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "-s",
    "--software",
    help="Archive software environments.",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "-r",
    "--results",
    help="Archive results files.",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "--compression",
    type=click.Choice(["none", "deflated", "bzip2", "lzma"], case_sensitive=False),
    default="none",
    help="Compression method.",
    show_default=True,
)
@click.option(
    "--compresslevel",
    type=int,
    default=None,
    help="Compression level.",
    show_default=True,
)
@click.option(
    "-n",
    "--dry-run",
    is_flag=True,
    default=False,
    help="Do not create the archive, just show what would be done.",
    show_default=True,
)
@click.option(
    "-l",
    "--local",
    help="Execute and store results locally. Default False.",
    is_flag=True,
    default=False,
)
@click.option(
    "--out-dir", type=str, default="out", help="Output folder name (local only)."
)
@click.pass_context
def archive_benchmark(
    ctx,
    benchmark,
    code,
    software,
    results,
    compression,
    compresslevel,
    dry_run,
    local,
    out_dir,
):
    # Validate out_dir usage
    if not local and out_dir != "out":
        logger.error("-Invalid arguments: --out-dir can only be used with --local")
        sys.exit(1)

    """Archive a benchmark"""
    benchmark = validate_benchmark(benchmark, "/tmp", echo=False)

    match compression:
        case "none":
            compression = zipfile.ZIP_STORED
        case "deflated":
            compression = zipfile.ZIP_DEFLATED
        case "bzip2":
            compression = zipfile.ZIP_BZIP2
        case "lzma":
            compression = zipfile.ZIP_LZMA
        case _:
            compression = zipfile.ZIP_STORED
    archive_file = archive_version(
        benchmark,
        outdir=Path("."),
        config=True,
        code=code,
        software=software,
        results=results,
        results_dir=out_dir,
        compression=compression,
        compresslevel=compresslevel,
        dry_run=dry_run,
        local=local,
    )
    if dry_run:
        click.echo(f"Files to archive:\n{tree_string_from_list(archive_file)}")
    else:
        click.echo(f"Created archive: {archive_file}")



"""Functions to manage files"""

import warnings
from pathlib import Path

import tqdm
import yaml

from omnibenchmark.benchmark import Benchmark
from omnibenchmark.cli.utils.logging import logger
from omnibenchmark.io.RemoteStorage import StorageOptions
from omnibenchmark.io.versioning import get_expected_benchmark_output_files


def list_files(
    benchmark_path: str,
    type: str,
    stage: str,
    module: str,
    file_id: str,
    local: bool = False,
    storage_options: StorageOptions = StorageOptions(out_dir="out")
):
    """List all available files for a certain benchmark, version and stage"""
    from .storage import get_storage, remote_storage_args

    with open(benchmark_path, "r") as fh:
        yaml.safe_load(fh)
        benchmark = Benchmark(Path(benchmark_path))

    expected_files = get_expected_benchmark_output_files(benchmark, storage_options)

    if not local:
        auth_options = remote_storage_args(benchmark)

        ss = get_storage(
            str(benchmark.converter.model.storage_api),
            auth_options,
            str(benchmark.converter.model.storage_bucket_name),
            storage_options
        )
        ss.set_version(benchmark.get_benchmark_version())
        ss._get_objects()
        files = {k: v for k, v in ss.files.items() if k in expected_files}

        # get urls
        objectnames = list(files.keys())
        etags = [files[objectname]["etag"] for objectname in objectnames]

    else:
        file_local_is_local = [Path(f).is_file() for f in expected_files]
        if not all(file_local_is_local):
            logger.warning("Not all expected files are available locally")
            logger.info("Missing files:")
            for i, f in enumerate(expected_files):
                if not file_local_is_local[i]:
                    logger.info(f)
        # make unique
        objectnames = list(set([Path(f) for f in expected_files]))
        etags = []

    return objectnames, etags


def download_files(
    benchmark_path: str,
    type: str,
    stage: str,
    module: str,
    file_id: str,
    verbose: bool = False,
    overwrite: bool = False,
):
    """Download all available files for a certain benchmark, version and stage"""
    from .storage import get_storage, remote_storage_args
    from .hash import checksum
    from .sizeof import sizeof_fmt

    objectnames, etags = list_files(benchmark_path, type, stage, module, file_id)

    # storage path locally, TODO: maybe add as argument
    filenames = objectnames

    with open(benchmark_path, "r") as fh:
        yaml.safe_load(fh)
        benchmark = Benchmark(Path(benchmark_path))

    auth_options = remote_storage_args(benchmark)

    ss = get_storage(
        str(benchmark.converter.model.storage_api),
        auth_options,
        str(benchmark.converter.model.storage_bucket_name),
    )
    ss.set_version(benchmark.get_benchmark_version())
    ss._get_objects()

    logger.debug("Checking if files are already downloaded... ")
    do_download_file = []
    for filename, etag in tqdm.tqdm(
        zip(filenames, etags), delay=5, disable=not verbose
    ):
        if Path(filename).is_file():
            if etag.replace('"', "") == checksum(filename.as_posix()):
                do_download_file.append(False)
            else:
                if overwrite:
                    do_download_file.append(True)
                else:
                    do_download_file.append(False)
        else:
            do_download_file.append(True)

    if verbose:
        size = sum(
            [
                int(ss.files[objectname]["size"])
                for objectname, do_download in zip(ss.files.keys(), do_download_file)
                if do_download
            ]
        )
        logger.debug(
            f"Downloading {sum(do_download_file)} files with a total size of {sizeof_fmt(size)} ... ",
        )

    for objectname, filename, do_download in tqdm.tqdm(
        zip(objectnames, filenames, do_download_file), delay=5, disable=not verbose
    ):
        if do_download:
            ss.download_object(objectname, filename)
    if verbose:
        logger.debug("Done")

    if verbose:
        logger.debug("Checking MD5 checksums... ")
    for etag, filename in tqdm.tqdm(
        zip(etags, filenames), delay=5, disable=not verbose
    ):
        md5sum_val = checksum(filename)
        if not etag.replace('"', "") == md5sum_val:
            warnings.warn(f"MD5 checksum failed for {filename}", Warning)
    if verbose:
        logger.debug("Done")

    return filenames


def checksum_files(
    benchmark: str,
    type: str,
    stage: str,
    module: str,
    file_id: str,
    verbose: bool = False,
):
    """Compare md5 checksums of available files for a certain benchmark, version and stage with local versions"""
    from .hash import checksum

    objectnames, etags = list_files(benchmark, type, stage, module, file_id)

    filenames = objectnames

    failed_checksums = []
    for etag, filename in tqdm.tqdm(
        zip(etags, filenames), delay=5, disable=not verbose
    ):
        md5sum_val = checksum(filename.as_posix())
        if not etag.replace('"', "") == md5sum_val:
            failed_checksums.append(filename)
    return failed_checksums



"""cli commands related to benchmark infos and stats"""

from packaging.version import Version
from pathlib import Path

import click
import yaml

from datetime import datetime
from difflib import unified_diff

from omnibenchmark.benchmark import Benchmark
from omnibenchmark.cli.utils.logging import logger
from omnibenchmark.cli.utils.validation import validate_benchmark
from omnibenchmark.io.storage import get_storage, remote_storage_args


@click.group(name="info")
@click.pass_context
# @debug_option
def info(ctx):
    """List benchmarks and/or information about them."""
    ctx.ensure_object(dict)


@info.command("diff")
@click.option(
    "--benchmark",
    "-b",
    required=True,
    type=click.Path(exists=True),
    help="Path to benchmark yaml file or benchmark id.",
    envvar="OB_BENCHMARK",
)
@click.option(
    "--version1",
    "-v1",
    required=True,
    type=str,
    help="Reference version.",
)
@click.option(
    "--version2",
    "-v2",
    required=True,
    type=str,
    help="Version to compare with.",
)
@click.pass_context
def diff_benchmark(ctx, benchmark, version1, version2):
    """Show differences between 2 benchmark versions."""
    logger.info(
        f"Found the following differences in {benchmark} for {version1} and {version2}."
    )
    with open(benchmark, "r") as fh:
        yaml.safe_load(fh)
        benchmark = Benchmark(Path(benchmark))

    auth_options = remote_storage_args(benchmark)

    # setup storage
    ss = get_storage(
        str(benchmark.converter.model.storage_api),
        auth_options,
        str(benchmark.converter.model.storage_bucket_name),
    )

    # get objects for first version
    ss.set_version(version1)
    ss._get_objects()
    files_v1 = [
        f"{f[0]}   {f[1]['size']}   {datetime.fromisoformat(f[1]['last_modified']).strftime('%Y-%m-%d %H:%M:%S')}\n"
        for f in ss.files.items()
    ]
    if f"versions/{version1}.csv" in ss.files.keys():
        creation_time_v1 = datetime.fromisoformat(
            ss.files[f"versions/{version1}.csv"]["last_modified"]
        ).strftime("%Y-%m-%d %H:%M:%S")
    else:
        creation_time_v1 = ""

    # get objects for second version
    ss.set_version(version2)
    ss._get_objects()
    files_v2 = [
        f"{f[0]}   {f[1]['size']}   {datetime.fromisoformat(f[1]['last_modified']).strftime('%Y-%m-%d %H:%M:%S')}\n"
        for f in ss.files.items()
    ]
    if f"versions/{version2}.csv" in ss.files.keys():
        creation_time_v2 = datetime.fromisoformat(
            ss.files[f"versions/{version2}.csv"]["last_modified"]
        ).strftime("%Y-%m-%d %H:%M:%S")
    else:
        creation_time_v2 = ""

    # diff the two versions
    click.echo(
        "".join(
            list(
                unified_diff(
                    files_v1,
                    files_v2,
                    fromfile=f"version {version1}",
                    tofile=f"version {version2}",
                    fromfiledate=creation_time_v1,
                    tofiledate=creation_time_v2,
                )
            )
        )
    )


@info.command("list-versions")
@click.option(
    "--benchmark",
    "-b",
    required=True,
    type=click.Path(exists=True),
    help="Path to benchmark yaml file or benchmark id.",
    envvar="OB_BENCHMARK",
)
@click.pass_context
def list_versions(ctx, benchmark):
    """List all available benchmarks versions at a specific endpoint."""
    logger.info(f"Available versions of {benchmark}:")

    with open(benchmark, "r") as fh:
        yaml.safe_load(fh)
        benchmark = Benchmark(Path(benchmark))

    auth_options = remote_storage_args(benchmark)

    # setup storage
    ss = get_storage(
        str(benchmark.converter.model.storage_api),
        auth_options,
        str(benchmark.converter.model.storage_bucket_name),
    )

    if len(ss.versions) > 0:
        if len(ss.versions) > 1:
            ss.versions.sort(key=Version)
        for version in ss.versions:
            click.echo(f"{str(version):>8}")


@info.command("computational")
@click.option(
    "--benchmark",
    "-b",
    required=True,
    type=click.Path(exists=True),
    help="Path to benchmark yaml file or benchmark id.",
    envvar="OB_BENCHMARK",
)
@click.pass_context
def computational_graph(ctx, benchmark: str):
    """Export computational graph to dot format."""

    b = validate_benchmark(benchmark, "/tmp", echo=False)
    if b is not None:
        dot = b.export_to_dot()
        click.echo(dot.to_string())


@info.command("topology")
@click.option(
    "--benchmark",
    "-b",
    required=True,
    type=click.Path(exists=True),
    help="Path to benchmark yaml file or benchmark id.",
    envvar="OB_BENCHMARK",
)
@click.pass_context
def plot_topology(ctx, benchmark: str):
    """Export benchmark topology to mermaid diagram format."""

    b = validate_benchmark(benchmark, "/tmp", echo=False)
    if b is not None:
        mermaid = b.export_to_mermaid()
        click.echo(mermaid)



from collections import defaultdict


def tree_string_from_list(file_list):
    file_tree = _tree()

    for file_path in file_list:
        parts = file_path.parts
        current_level = file_tree
        for part in parts:
            current_level = current_level[part]

    def print_tree(current_level, indent=""):
        out = ""
        for key, subtree in current_level.items():
            out += f"{indent}{key}/\n"
            out += print_tree(subtree, indent + "    ")
        return out

    return print_tree(file_tree)


def _tree():
    """
    A recursive function that creates a tree structure from a list of file paths.
    """
    return defaultdict(_tree)



import os.path
from pathlib import Path
from abc import ABCMeta, abstractmethod
from typing import Optional

from omni_schema.datamodel.omni_schema import SoftwareBackendEnum

from omnibenchmark.benchmark import Benchmark, BenchmarkNode


class WorkflowEngine(metaclass=ABCMeta):
    """Interface for the workflow engine."""

    @abstractmethod
    def run_workflow(
        self,
        benchmark: Benchmark,
        cores: int = 1,
        update: bool = True,
        dryrun: bool = False,
        continue_on_error: bool = False,
        keep_module_logs: bool = False,
        backend: SoftwareBackendEnum = SoftwareBackendEnum.host,
        module_path: str = os.environ.get("MODULEPATH", None),
        local_timeout: Optional[int] = None,
        work_dir: Path = Path(os.getcwd()),
        out_dir: str = "out",
        resources: Optional[dict] = None,
        **kwargs,
    ) -> bool:
        """
        Serializes & runs benchmark workflow.

        Args:
            benchmark (Benchmark): benchmark to run
            cores (int): number of cores to run. Defaults to 1 core.
            update (bool): run workflow for non-existing outputs / changed nodes only. False means force running workflow from scratch. Default: True
            dryrun (bool): validate the workflow with the benchmark without actual execution. Default: False
            continue_on_error (bool): continue with independent jobs if a job fails. Default: False
            keep_module_logs (bool): keep module-specific log files after execution. Default: False
            backend (SoftwareBackendEnum): which software backend to use when running the workflow. Available: `host`, `docker`, `apptainer`, `conda`, `envmodules`. Default: `host`
            module_path (str): The path where the `envmodules` are located. This path will be searched during the workflow run using `envmodules` backend.
            work_dir (str): working directory. Default: current work directory
            out_dir (str): output directory. Default: `out`
            local_timeout (int, optional): timeout, in seconds, when executing locally. It will be passed to the subprocess invocation.
            **kwargs: keyword arguments to pass to the workflow engine

        Returns:
        - Status code (bool) of the workflow run.
        """
        raise NotImplementedError("Method not implemented yet")

    @abstractmethod
    def serialize_workflow(
        self, benchmark: Benchmark, output_dir: Path = Path(os.getcwd())
    ) -> Path:
        """
        Serializes a workflow file for the benchmark.

        Args:
            benchmark (Benchmark): benchmark to serialize
            output_dir (str): output directory for the workflow file

        Returns:
        - Workflow file path.
        """
        raise NotImplementedError("Method not implemented yet")

    @abstractmethod
    def run_node_workflow(
        self,
        node: BenchmarkNode,
        input_dir: Path,
        dataset: str,
        cores: int = 1,
        update: bool = True,
        dryrun: bool = False,
        continue_on_error: bool = False,
        keep_module_logs: bool = False,
        backend: SoftwareBackendEnum = SoftwareBackendEnum.host,
        module_path: str = os.environ.get("MODULEPATH", None),
        work_dir: Path = Path(os.getcwd()),
        **kwargs,
    ) -> bool:
        """
        Serializes & runs benchmark node workflow.

        Args:
            node (Benchmark): benchmark node to run
            input_dir (str): directory containing the inputs for the benchmark node
            dataset (str): file names corresponding to the dataset
            cores (int): number of cores to run. Defaults to 1 core.
            update (bool): run workflow for non-existing outputs / changed nodes only. False means force running workflow from scratch. Default: True
            dryrun (bool): validate the workflow with the benchmark without actual execution. Default: False
            continue_on_error (bool): continue with independent jobs if a job fails. Default: False
            keep_module_logs (bool): keep module-specific log files after execution. Default: False
            backend (SoftwareBackendEnum): which software backend to use when running the workflow. Available: `host`, `docker`, `apptainer`, `conda`, `envmodules`. Default: `host`
            module_path (str): The path where the `envmodules` are located. This path will be searched during the workflow run using `envmodules` backend.
            work_dir (str): working directory. Default: current work directory
            **kwargs: keyword arguments to pass to the workflow engine

        Returns:
        - Status code (bool) of the workflow run.
        """
        raise NotImplementedError("Method not implemented yet")

    @abstractmethod
    def serialize_node_workflow(
        self, node: BenchmarkNode, output_dir: Path = Path(os.getcwd())
    ) -> Path:
        """
        Serializes a workflow file for a benchmark node.

        Args:
            node (BenchmarkNode): benchmark node to serialize
            output_dir (str): output directory for the workflow file

        Returns:
        - Workflow file path.
        """
        raise NotImplementedError("Method not implemented yet")
